import json
import logging
from typing import List

from semgrep.app import app_session
from semgrep.constants import SEMGREP_URL
from semgrep.error import SemgrepError
from semgrep.types import JsonObject

logger = logging.getLogger(__name__)


def list_current_public_rulesets() -> List[JsonObject]:
    api_full_url = f"{SEMGREP_URL}/api/registry/ruleset"
    try:
        r = app_session.get(api_full_url)
    except Exception:
        raise SemgrepError(f"Failed to download list of public rulesets")

    if not r.ok:
        raise SemgrepError(
            f"Bad status code: {r.status_code} returned by url: {api_full_url}"
        )

    logger.debug(f"Retrieved rulesets: {r.text}")
    try:
        ruleset_json = json.loads(r.text)
    except json.decoder.JSONDecodeError as e:
        raise SemgrepError(f"Failed to parse rulesets as valid json")

    return ruleset_json  # type:ignore



from pathlib import Path
from typing import Dict
from typing import Generator
from typing import List
from typing import Tuple

from dependencyparser.find_lockfiles import DependencyTrie
from dependencyparser.models import languages_to_namespaces
from dependencyparser.models import LockfileDependency
from dependencyparser.models import PackageManagers
from dependencyparser.package_restrictions import dependencies_range_match_any
from dependencyparser.package_restrictions import ProjectDependsOnEntry
from semgrep.error import SemgrepError
from semgrep.output_from_core import Position
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch


PACKAGE_MANAGER_MAP = {
    "pypi": PackageManagers.PYPI,
    "npm": PackageManagers.NPM,
    "gem": PackageManagers.GEM,
    "gomod": PackageManagers.GOMOD,
    "cargo": PackageManagers.CARGO,
    "maven": PackageManagers.MAVEN,
}


def parse_depends_on_yaml(
    entries: List[Dict[str, str]]
) -> Generator[ProjectDependsOnEntry, None, None]:
    """
    Convert the entries in the Yaml to ProjectDependsOnEntry objects that specify
    namespace, package name, and semver ranges
    """
    for entry in entries:
        # schema checks should gaurantee we have these fields, but we'll code defensively
        namespace = entry.get("namespace")
        if namespace is None:
            raise SemgrepError(f"project-depends-on is missing `namespace`")
        pm = PACKAGE_MANAGER_MAP.get(namespace)
        if pm is None:
            raise SemgrepError(
                f"unknown package namespace: {namespace}, only {list(PACKAGE_MANAGER_MAP.keys())} are supported"
            )
        package_name = entry.get("package")
        if package_name is None:
            raise SemgrepError(f"project-depends-on is missing `package`")
        semver_range = entry.get("version")
        if semver_range is None:
            raise SemgrepError(f"project-depends-on is missing `version`")
        yield ProjectDependsOnEntry(
            namespace=pm, package_name=package_name, semver_range=semver_range
        )


def run_dependency_aware_rule(
    matches: List[RuleMatch],
    rule: Rule,
    dep_trie: DependencyTrie,
) -> Tuple[List[RuleMatch], List[SemgrepError]]:
    """
    Run a dependency aware rule. These rules filters the results based on the precense or absence
    of dependencies. Dependencies are determined by searching for lockfiles under the first entry
    in the `targets` argument.
    """
    depends_on_keys: List[Dict[str, str]] = rule.project_depends_on or []
    dep_rule_errors: List[SemgrepError] = []

    if len(depends_on_keys) == 0:
        # no dependencies to process, so skip
        return matches, []

    depends_on_entries = list(parse_depends_on_yaml(depends_on_keys))
    final_matches: List[RuleMatch] = []

    namespaces = languages_to_namespaces(rule.languages)
    dependencies: List[Tuple[Path, List[LockfileDependency]]] = []
    for ns in namespaces:
        if ns in dep_trie.all_deps:
            dependencies.extend(dep_trie.all_deps[ns].items())

    for lockfile_path, deps in dependencies:
        try:
            output = list(
                dependencies_range_match_any(depends_on_entries, lockfile_path, deps)
            )
            if not output:
                continue

            output_for_json = [
                {
                    "dependency_pattern": vars(dep_pat),
                    "found_dependency": vars(found_dep),
                    "lockfile": lockfile.name,
                }
                for dep_pat, found_dep, lockfile in output
            ]

            reachable = []
            matches_remaining = []
            for match in matches:
                if lockfile_path in (dep_trie.find_dependencies(match.path) or {}):
                    match.extra["dependency_match_only"] = False
                    match.extra["dependency_matches"] = output_for_json
                    reachable.append(match)
                else:
                    matches_remaining.append(match)
            matches = matches_remaining
            if not reachable:
                dependency_only_match = RuleMatch(
                    rule_id=rule.id,
                    message=rule.message,
                    metadata=rule.metadata,
                    severity=rule.severity,
                    path=lockfile_path,
                    fix=None,
                    fix_regex=None,
                    start=Position(0, 0, 0),
                    end=Position(0, 0, 0),
                    extra={
                        "dependency_match_only": True,
                        "dependency_matches": output_for_json,
                    },
                )
                final_matches.append(dependency_only_match)
            else:
                final_matches.extend(reachable)

        except SemgrepError as e:
            dep_rule_errors.append(e)
    return final_matches, dep_rule_errors



# Handle communication of findings / errors to semgrep.app
import json
import os
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

import click
import requests

from semgrep.app import app_session
from semgrep.constants import SEMGREP_URL
from semgrep.error import SemgrepError
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatchMap
from semgrep.util import partition
from semgrep.verbose_logging import getLogger


logger = getLogger(__name__)


class ScanHandler:
    def __init__(self, dry_run: bool) -> None:
        self.deployment_id, self.deployment_name = self._get_deployment_details()

        self.scan_id = None
        self.ignore_patterns: List[str] = []
        self._autofix = False
        self.dry_run = dry_run
        self._dry_run_rules_url: str = ""

    @property
    def autofix(self) -> bool:
        """
        Seperate property for easy of mocking in test
        """
        return self._autofix

    def _get_deployment_details(self) -> Tuple[Optional[int], Optional[str]]:
        """
        Returns the deployment_id attached to an api_token as int

        Returns None if api_token is invalid/doesn't have associated deployment
        """
        url = f"{SEMGREP_URL}/api/agent/deployments/current"
        logger.debug(f"Retrieving deployment details from {url}")
        r = app_session.get(url)

        if r.ok:
            data = r.json()
            logger.debug(f"Received: {data}")
            return data.get("deployment", {}).get("id"), data.get("deployment", {}).get(
                "name"
            )
        else:
            return None, None

    def start_scan(self, meta: Dict[str, Any]) -> None:
        """
        Get scan id and file ignores

        returns ignored list
        """
        logger.debug("Starting scan")
        if self.dry_run:
            repo_name = meta["repository"]
            self._dry_run_rules_url = f"{SEMGREP_URL}/api/agent/deployments/{self.deployment_id}/repos/{repo_name}/rules.yaml"
            logger.debug(
                f"ran with dryrun so setting rules url to {self._dry_run_rules_url}"
            )
            return

        response = app_session.post(
            f"{SEMGREP_URL}/api/agent/deployments/{self.deployment_id}/scans",
            json={"meta": meta},
        )

        if response.status_code == 404:
            raise Exception(
                "Failed to create a scan with given token and deployment_id."
                "Please make sure they have been set correctly."
                f"API server at {SEMGREP_URL} returned this response: {response.text}"
            )

        try:
            response.raise_for_status()
        except requests.RequestException:
            raise Exception(
                f"API server at {SEMGREP_URL} returned this error: {response.text}"
            )

        body = response.json()
        self.scan_id = body["scan"]["id"]
        self._autofix = body.get("autofix", False)
        self.ignore_patterns = body["scan"]["meta"].get("ignored_files", [])

    @property
    def scan_rules_url(self) -> str:
        if self.dry_run:
            url = self._dry_run_rules_url
        else:
            url = f"{SEMGREP_URL}/api/agent/scans/{self.scan_id}/rules.yaml"

        logger.debug(f"Using {url} as scan rules url")
        return url

    def report_failure(self, exit_code: int) -> None:
        """
        Send semgrep cli non-zero exit code information to server
        and return what exit code semgrep should exit with.
        """
        if self.dry_run:
            logger.info(f"Would have reported failure to semgrep.dev: {exit_code}")
            return

        response = app_session.post(
            f"{SEMGREP_URL}/api/agent/scans/{self.scan_id}/error",
            json={
                "exit_code": exit_code,
                "stderr": "",
            },
        )

        try:
            response.raise_for_status()
        except requests.RequestException:
            raise Exception(f"API server returned this error: {response.text}")

    def report_findings(
        self,
        matches_by_rule: RuleMatchMap,
        errors: List[SemgrepError],
        rules: List[Rule],
        targets: Set[Path],
        total_time: float,
        commit_date: str,
    ) -> None:
        """
        commit_date here for legacy reasons. epoch time of latest commit
        """
        all_ids = [r.id for r in rules]
        cai_ids, rule_ids = partition(
            lambda r_id: "r2c-internal-cai" in r_id,
            all_ids,
        )

        all_matches = [
            match
            for matches_of_rule in matches_by_rule.values()
            for match in matches_of_rule
        ]
        new_ignored, new_matches = partition(
            lambda match: match.is_ignored,
            all_matches,
        )

        findings = {
            # send a backup token in case the app is not available
            "token": os.getenv("GITHUB_TOKEN"),
            "gitlab_token": os.getenv("GITLAB_TOKEN"),
            "findings": [
                match.to_app_finding_format(commit_date) for match in new_matches
            ],
            "searched_paths": [str(t) for t in targets],
            "rule_ids": rule_ids,
            "cai_ids": cai_ids,
        }
        ignores = {
            "findings": [
                match.to_app_finding_format(commit_date) for match in new_ignored
            ],
        }
        complete = {
            "exit_code": 1
            if any(match.is_blocking and not match.is_ignored for match in all_matches)
            else 0,
            "stats": {
                "findings": len(new_matches),
                "errors": [error.to_dict() for error in errors],
                "total_time": total_time,
            },
        }

        if self.dry_run:
            logger.info(
                f"Would have sent findings blob: {json.dumps(findings, indent=4)}"
            )
            logger.info(
                f"Would have sent ignores blob: {json.dumps(ignores, indent=4)}"
            )
            logger.info(
                f"Would have sent complete blob: {json.dumps(complete, indent=4)}"
            )
            return
        else:
            logger.debug(f"Sending findings blob: {json.dumps(findings, indent=4)}")
            logger.debug(f"Sending ignores blob: {json.dumps(ignores, indent=4)}")
            logger.debug(f"Sending complete blob: {json.dumps(complete, indent=4)}")

        response = app_session.post(
            f"{SEMGREP_URL}/api/agent/scans/{self.scan_id}/findings", json=findings
        )
        try:
            response.raise_for_status()

            resp_errors = response.json()["errors"]
            for error in resp_errors:
                message = error["message"]
                click.echo(f"Server returned following warning: {message}", err=True)

        except requests.RequestException:
            raise Exception(f"API server returned this error: {response.text}")

        response = app_session.post(
            f"{SEMGREP_URL}/api/agent/scans/{self.scan_id}/ignores", json=ignores
        )
        try:
            response.raise_for_status()
        except requests.RequestException:
            raise Exception(f"API server returned this error: {response.text}")

        # mark as complete
        response = app_session.post(
            f"{SEMGREP_URL}/api/agent/scans/{self.scan_id}/complete", json=complete
        )

        try:
            response.raise_for_status()
        except requests.RequestException:
            raise Exception(
                f"API server at {SEMGREP_URL} returned this error: {response.text}"
            )



from dataclasses import dataclass
from pathlib import Path
from typing import Generator
from typing import List
from typing import Tuple

import packaging.version
from packaging.specifiers import SpecifierSet

from dependencyparser.models import LockfileDependency
from dependencyparser.models import PackageManagers
from semgrep.error import SemgrepError


@dataclass(eq=True, order=True, frozen=True)
class ProjectDependsOnEntry:
    namespace: PackageManagers
    package_name: str
    semver_range: str


def semver_matches(expression: str, actual_version: str) -> bool:

    try:
        ss = SpecifierSet(expression)
        matched = len(list(ss.filter([actual_version]))) > 0
        # print(f'does {expression} match {actual_version}?: {matched}')
        return matched
    except packaging.specifiers.InvalidSpecifier:
        raise SemgrepError(
            f"unknown package version comparison expression: {expression}"
        )


# compare vulnerable range to version in lockfile
def dependencies_range_match_any(
    search_for_ranges: List[ProjectDependsOnEntry],
    lockfile_path: Path,
    have_deps: List[LockfileDependency],
) -> Generator[Tuple[ProjectDependsOnEntry, LockfileDependency, Path], None, None]:
    for have_dep in have_deps:
        for target_range in search_for_ranges:
            if (
                target_range.namespace.value.lower() == have_dep.namespace.value.lower()
                and target_range.package_name == have_dep.name
                and semver_matches(target_range.semver_range, have_dep.version)
            ):
                yield (target_range, have_dep, lockfile_path)



"""Generated by atdpy from type definitions in Output_from_core.atd.

This implements classes for the types defined in 'Output_from_core.atd', providing
methods and functions to convert data from/to JSON.
"""
# Disable flake8 entirely on this file:
# flake8: noqa
import json
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import NoReturn
from typing import Optional
from typing import Tuple
from typing import Union

############################################################################
# Private functions
############################################################################


def _atd_missing_json_field(type_name: str, json_field_name: str) -> NoReturn:
    raise ValueError(
        f"missing field '{json_field_name}'" f" in JSON object of type '{type_name}'"
    )


def _atd_bad_json(expected_type: str, json_value: Any) -> NoReturn:
    value_str = str(json_value)
    if len(value_str) > 200:
        value_str = value_str[:200] + "…"

    raise ValueError(
        f"incompatible JSON value where"
        f" type '{expected_type}' was expected: '{value_str}'"
    )


def _atd_bad_python(expected_type: str, json_value: Any) -> NoReturn:
    value_str = str(json_value)
    if len(value_str) > 200:
        value_str = value_str[:200] + "…"

    raise ValueError(
        f"incompatible Python value where"
        f" type '{expected_type}' was expected: '{value_str}'"
    )


def _atd_read_unit(x: Any) -> None:
    if x is None:
        return x
    else:
        _atd_bad_json("unit", x)


def _atd_read_bool(x: Any) -> bool:
    if isinstance(x, bool):
        return x
    else:
        _atd_bad_json("bool", x)


def _atd_read_int(x: Any) -> int:
    if isinstance(x, int):
        return x
    else:
        _atd_bad_json("int", x)


def _atd_read_float(x: Any) -> float:
    if isinstance(x, (int, float)):
        return x
    else:
        _atd_bad_json("float", x)


def _atd_read_string(x: Any) -> str:
    if isinstance(x, str):
        return x
    else:
        _atd_bad_json("str", x)


def _atd_read_list(read_elt: Callable[[Any], Any]) -> Callable[[List[Any]], List[Any]]:
    def read_list(elts: List[Any]) -> List[Any]:
        if isinstance(elts, list):
            return [read_elt(elt) for elt in elts]
        else:
            _atd_bad_json("array", elts)

    return read_list


def _atd_read_assoc_array_into_dict(
    read_key: Callable[[Any], Any],
    read_value: Callable[[Any], Any],
) -> Callable[[List[Any]], Dict[Any, Any]]:
    def read_assoc(elts: List[List[Any]]) -> Dict[str, Any]:
        if isinstance(elts, list):
            return {read_key(elt[0]): read_value(elt[1]) for elt in elts}
        else:
            _atd_bad_json("array", elts)
            raise AssertionError("impossible")  # keep mypy happy

    return read_assoc


def _atd_read_assoc_object_into_dict(
    read_value: Callable[[Any], Any]
) -> Callable[[Dict[str, Any]], Dict[str, Any]]:
    def read_assoc(elts: Dict[str, Any]) -> Dict[str, Any]:
        if isinstance(elts, dict):
            return {_atd_read_string(k): read_value(v) for k, v in elts.items()}
        else:
            _atd_bad_json("object", elts)
            raise AssertionError("impossible")  # keep mypy happy

    return read_assoc


def _atd_read_assoc_object_into_list(
    read_value: Callable[[Any], Any]
) -> Callable[[Dict[str, Any]], List[Tuple[str, Any]]]:
    def read_assoc(elts: Dict[str, Any]) -> List[Tuple[str, Any]]:
        if isinstance(elts, dict):
            return [(_atd_read_string(k), read_value(v)) for k, v in elts.items()]
        else:
            _atd_bad_json("object", elts)
            raise AssertionError("impossible")  # keep mypy happy

    return read_assoc


def _atd_read_nullable(
    read_elt: Callable[[Any], Any]
) -> Callable[[Optional[Any]], Optional[Any]]:
    def read_nullable(x: Any) -> Any:
        if x is None:
            return None
        else:
            return read_elt(x)

    return read_nullable


def _atd_write_unit(x: Any) -> None:
    if x is None:
        return x
    else:
        _atd_bad_python("unit", x)


def _atd_write_bool(x: Any) -> bool:
    if isinstance(x, bool):
        return x
    else:
        _atd_bad_python("bool", x)


def _atd_write_int(x: Any) -> int:
    if isinstance(x, int):
        return x
    else:
        _atd_bad_python("int", x)


def _atd_write_float(x: Any) -> float:
    if isinstance(x, (int, float)):
        return x
    else:
        _atd_bad_python("float", x)


def _atd_write_string(x: Any) -> str:
    if isinstance(x, str):
        return x
    else:
        _atd_bad_python("str", x)


def _atd_write_list(
    write_elt: Callable[[Any], Any]
) -> Callable[[List[Any]], List[Any]]:
    def write_list(elts: List[Any]) -> List[Any]:
        if isinstance(elts, list):
            return [write_elt(elt) for elt in elts]
        else:
            _atd_bad_python("list", elts)

    return write_list


def _atd_write_assoc_dict_to_array(
    write_key: Callable[[Any], Any], write_value: Callable[[Any], Any]
) -> Callable[[Dict[Any, Any]], List[Tuple[Any, Any]]]:
    def write_assoc(elts: Dict[str, Any]) -> List[Tuple[str, Any]]:
        if isinstance(elts, dict):
            return [(write_key(k), write_value(v)) for k, v in elts.items()]
        else:
            _atd_bad_python("Dict[str, <value type>]]", elts)
            raise AssertionError("impossible")  # keep mypy happy

    return write_assoc


def _atd_write_assoc_dict_to_object(
    write_value: Callable[[Any], Any]
) -> Callable[[Dict[str, Any]], Dict[str, Any]]:
    def write_assoc(elts: Dict[str, Any]) -> Dict[str, Any]:
        if isinstance(elts, dict):
            return {_atd_write_string(k): write_value(v) for k, v in elts.items()}
        else:
            _atd_bad_python("Dict[str, <value type>]", elts)
            raise AssertionError("impossible")  # keep mypy happy

    return write_assoc


def _atd_write_assoc_list_to_object(
    write_value: Callable[[Any], Any],
) -> Callable[[List[Any]], Dict[str, Any]]:
    def write_assoc(elts: List[List[Any]]) -> Dict[str, Any]:
        if isinstance(elts, list):
            return {_atd_write_string(elt[0]): write_value(elt[1]) for elt in elts}
        else:
            _atd_bad_python("List[Tuple[<key type>, <value type>]]", elts)
            raise AssertionError("impossible")  # keep mypy happy

    return write_assoc


def _atd_write_nullable(
    write_elt: Callable[[Any], Any]
) -> Callable[[Optional[Any]], Optional[Any]]:
    def write_nullable(x: Any) -> Any:
        if x is None:
            return None
        else:
            return write_elt(x)

    return write_nullable


############################################################################
# Public classes
############################################################################


@dataclass
class ID:
    """Original type: unique_id_type = [ ... | ID | ... ]"""

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return "ID"

    @staticmethod
    def to_json() -> Any:
        return "id"

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass
class AST:
    """Original type: unique_id_type = [ ... | AST | ... ]"""

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return "AST"

    @staticmethod
    def to_json() -> Any:
        return "AST"

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class UniqueIdType:
    """Original type: unique_id_type = [ ... ]"""

    value: Union[ID, AST]

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return self.value.kind

    @classmethod
    def from_json(cls, x: Any) -> "UniqueIdType":
        if isinstance(x, str):
            if x == "id":
                return cls(ID())
            if x == "AST":
                return cls(AST())
            _atd_bad_json("UniqueIdType", x)
        _atd_bad_json("UniqueIdType", x)

    def to_json(self) -> Any:
        return self.value.to_json()

    @classmethod
    def from_json_string(cls, x: str) -> "UniqueIdType":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class UniqueId:
    """Original type: unique_id = { ... }"""

    type_: UniqueIdType
    md5sum: Optional[str] = None
    sid: Optional[int] = None

    @classmethod
    def from_json(cls, x: Any) -> "UniqueId":
        if isinstance(x, dict):
            return cls(
                type_=UniqueIdType.from_json(x["type"])
                if "type" in x
                else _atd_missing_json_field("UniqueId", "type"),
                md5sum=_atd_read_string(x["md5sum"]) if "md5sum" in x else None,
                sid=_atd_read_int(x["sid"]) if "sid" in x else None,
            )
        else:
            _atd_bad_json("UniqueId", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["type"] = (lambda x: x.to_json())(self.type_)
        if self.md5sum is not None:
            res["md5sum"] = _atd_write_string(self.md5sum)
        if self.sid is not None:
            res["sid"] = _atd_write_int(self.sid)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "UniqueId":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class RuleTimes:
    """Original type: rule_times = { ... }"""

    rule_id: str
    parse_time: float
    match_time: float

    @classmethod
    def from_json(cls, x: Any) -> "RuleTimes":
        if isinstance(x, dict):
            return cls(
                rule_id=_atd_read_string(x["rule_id"])
                if "rule_id" in x
                else _atd_missing_json_field("RuleTimes", "rule_id"),
                parse_time=_atd_read_float(x["parse_time"])
                if "parse_time" in x
                else _atd_missing_json_field("RuleTimes", "parse_time"),
                match_time=_atd_read_float(x["match_time"])
                if "match_time" in x
                else _atd_missing_json_field("RuleTimes", "match_time"),
            )
        else:
            _atd_bad_json("RuleTimes", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["rule_id"] = _atd_write_string(self.rule_id)
        res["parse_time"] = _atd_write_float(self.parse_time)
        res["match_time"] = _atd_write_float(self.match_time)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "RuleTimes":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class TargetTime:
    """Original type: target_time = { ... }"""

    path: str
    rule_times: List[RuleTimes]
    run_time: float

    @classmethod
    def from_json(cls, x: Any) -> "TargetTime":
        if isinstance(x, dict):
            return cls(
                path=_atd_read_string(x["path"])
                if "path" in x
                else _atd_missing_json_field("TargetTime", "path"),
                rule_times=_atd_read_list(RuleTimes.from_json)(x["rule_times"])
                if "rule_times" in x
                else _atd_missing_json_field("TargetTime", "rule_times"),
                run_time=_atd_read_float(x["run_time"])
                if "run_time" in x
                else _atd_missing_json_field("TargetTime", "run_time"),
            )
        else:
            _atd_bad_json("TargetTime", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["path"] = _atd_write_string(self.path)
        res["rule_times"] = _atd_write_list(lambda x: x.to_json())(self.rule_times)
        res["run_time"] = _atd_write_float(self.run_time)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "TargetTime":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class Time:
    """Original type: time = { ... }"""

    targets: List[TargetTime]
    rules: List[str]
    rules_parse_time: Optional[float] = None

    @classmethod
    def from_json(cls, x: Any) -> "Time":
        if isinstance(x, dict):
            return cls(
                targets=_atd_read_list(TargetTime.from_json)(x["targets"])
                if "targets" in x
                else _atd_missing_json_field("Time", "targets"),
                rules=_atd_read_list(_atd_read_string)(x["rules"])
                if "rules" in x
                else _atd_missing_json_field("Time", "rules"),
                rules_parse_time=_atd_read_float(x["rules_parse_time"])
                if "rules_parse_time" in x
                else None,
            )
        else:
            _atd_bad_json("Time", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["targets"] = _atd_write_list(lambda x: x.to_json())(self.targets)
        res["rules"] = _atd_write_list(_atd_write_string)(self.rules)
        if self.rules_parse_time is not None:
            res["rules_parse_time"] = _atd_write_float(self.rules_parse_time)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "Time":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class Stats:
    """Original type: stats = { ... }"""

    okfiles: int
    errorfiles: int

    @classmethod
    def from_json(cls, x: Any) -> "Stats":
        if isinstance(x, dict):
            return cls(
                okfiles=_atd_read_int(x["okfiles"])
                if "okfiles" in x
                else _atd_missing_json_field("Stats", "okfiles"),
                errorfiles=_atd_read_int(x["errorfiles"])
                if "errorfiles" in x
                else _atd_missing_json_field("Stats", "errorfiles"),
            )
        else:
            _atd_bad_json("Stats", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["okfiles"] = _atd_write_int(self.okfiles)
        res["errorfiles"] = _atd_write_int(self.errorfiles)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "Stats":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass
class ExcludedByConfig:
    """Original type: skip_reason = [ ... | Excluded_by_config | ... ]"""

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return "ExcludedByConfig"

    @staticmethod
    def to_json() -> Any:
        return "excluded_by_config"

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass
class WrongLanguage:
    """Original type: skip_reason = [ ... | Wrong_language | ... ]"""

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return "WrongLanguage"

    @staticmethod
    def to_json() -> Any:
        return "wrong_language"

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass
class TooBig:
    """Original type: skip_reason = [ ... | Too_big | ... ]"""

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return "TooBig"

    @staticmethod
    def to_json() -> Any:
        return "too_big"

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass
class Minified:
    """Original type: skip_reason = [ ... | Minified | ... ]"""

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return "Minified"

    @staticmethod
    def to_json() -> Any:
        return "minified"

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass
class Binary:
    """Original type: skip_reason = [ ... | Binary | ... ]"""

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return "Binary"

    @staticmethod
    def to_json() -> Any:
        return "binary"

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass
class IrrelevantRule:
    """Original type: skip_reason = [ ... | Irrelevant_rule | ... ]"""

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return "IrrelevantRule"

    @staticmethod
    def to_json() -> Any:
        return "irrelevant_rule"

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass
class TooManyMatches:
    """Original type: skip_reason = [ ... | Too_many_matches | ... ]"""

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return "TooManyMatches"

    @staticmethod
    def to_json() -> Any:
        return "too_many_matches"

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class SkipReason:
    """Original type: skip_reason = [ ... ]"""

    value: Union[
        ExcludedByConfig,
        WrongLanguage,
        TooBig,
        Minified,
        Binary,
        IrrelevantRule,
        TooManyMatches,
    ]

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return self.value.kind

    @classmethod
    def from_json(cls, x: Any) -> "SkipReason":
        if isinstance(x, str):
            if x == "excluded_by_config":
                return cls(ExcludedByConfig())
            if x == "wrong_language":
                return cls(WrongLanguage())
            if x == "too_big":
                return cls(TooBig())
            if x == "minified":
                return cls(Minified())
            if x == "binary":
                return cls(Binary())
            if x == "irrelevant_rule":
                return cls(IrrelevantRule())
            if x == "too_many_matches":
                return cls(TooManyMatches())
            _atd_bad_json("SkipReason", x)
        _atd_bad_json("SkipReason", x)

    def to_json(self) -> Any:
        return self.value.to_json()

    @classmethod
    def from_json_string(cls, x: str) -> "SkipReason":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class RuleId:
    """Original type: rule_id"""

    value: str

    @classmethod
    def from_json(cls, x: Any) -> "RuleId":
        return cls(_atd_read_string(x))

    def to_json(self) -> Any:
        return _atd_write_string(self.value)

    @classmethod
    def from_json_string(cls, x: str) -> "RuleId":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class SkippedTarget:
    """Original type: skipped_target = { ... }"""

    path: str
    reason: SkipReason
    details: str
    rule_id: Optional[RuleId] = None

    @classmethod
    def from_json(cls, x: Any) -> "SkippedTarget":
        if isinstance(x, dict):
            return cls(
                path=_atd_read_string(x["path"])
                if "path" in x
                else _atd_missing_json_field("SkippedTarget", "path"),
                reason=SkipReason.from_json(x["reason"])
                if "reason" in x
                else _atd_missing_json_field("SkippedTarget", "reason"),
                details=_atd_read_string(x["details"])
                if "details" in x
                else _atd_missing_json_field("SkippedTarget", "details"),
                rule_id=RuleId.from_json(x["rule_id"]) if "rule_id" in x else None,
            )
        else:
            _atd_bad_json("SkippedTarget", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["path"] = _atd_write_string(self.path)
        res["reason"] = (lambda x: x.to_json())(self.reason)
        res["details"] = _atd_write_string(self.details)
        if self.rule_id is not None:
            res["rule_id"] = (lambda x: x.to_json())(self.rule_id)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "SkippedTarget":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True, order=True)
class Position:
    """Original type: position = { ... }"""

    line: int
    col: int
    offset: int

    @classmethod
    def from_json(cls, x: Any) -> "Position":
        if isinstance(x, dict):
            return cls(
                line=_atd_read_int(x["line"])
                if "line" in x
                else _atd_missing_json_field("Position", "line"),
                col=_atd_read_int(x["col"])
                if "col" in x
                else _atd_missing_json_field("Position", "col"),
                offset=_atd_read_int(x["offset"])
                if "offset" in x
                else _atd_missing_json_field("Position", "offset"),
            )
        else:
            _atd_bad_json("Position", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["line"] = _atd_write_int(self.line)
        res["col"] = _atd_write_int(self.col)
        res["offset"] = _atd_write_int(self.offset)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "Position":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class SkippedRule:
    """Original type: skipped_rule = { ... }"""

    rule_id: RuleId
    details: str
    position: Position

    @classmethod
    def from_json(cls, x: Any) -> "SkippedRule":
        if isinstance(x, dict):
            return cls(
                rule_id=RuleId.from_json(x["rule_id"])
                if "rule_id" in x
                else _atd_missing_json_field("SkippedRule", "rule_id"),
                details=_atd_read_string(x["details"])
                if "details" in x
                else _atd_missing_json_field("SkippedRule", "details"),
                position=Position.from_json(x["position"])
                if "position" in x
                else _atd_missing_json_field("SkippedRule", "position"),
            )
        else:
            _atd_bad_json("SkippedRule", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["rule_id"] = (lambda x: x.to_json())(self.rule_id)
        res["details"] = _atd_write_string(self.details)
        res["position"] = (lambda x: x.to_json())(self.position)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "SkippedRule":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass
class Error_:
    """Original type: severity = [ ... | Error | ... ]"""

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return "Error_"

    @staticmethod
    def to_json() -> Any:
        return "error"

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass
class Warning:
    """Original type: severity = [ ... | Warning | ... ]"""

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return "Warning"

    @staticmethod
    def to_json() -> Any:
        return "warning"

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class Severity:
    """Original type: severity = [ ... ]"""

    value: Union[Error_, Warning]

    @property
    def kind(self) -> str:
        """Name of the class representing this variant."""
        return self.value.kind

    @classmethod
    def from_json(cls, x: Any) -> "Severity":
        if isinstance(x, str):
            if x == "error":
                return cls(Error_())
            if x == "warning":
                return cls(Warning())
            _atd_bad_json("Severity", x)
        _atd_bad_json("Severity", x)

    def to_json(self) -> Any:
        return self.value.to_json()

    @classmethod
    def from_json_string(cls, x: str) -> "Severity":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class MetavarValue:
    """Original type: metavar_value = { ... }"""

    start: Position
    end: Position
    abstract_content: str
    unique_id: UniqueId

    @classmethod
    def from_json(cls, x: Any) -> "MetavarValue":
        if isinstance(x, dict):
            return cls(
                start=Position.from_json(x["start"])
                if "start" in x
                else _atd_missing_json_field("MetavarValue", "start"),
                end=Position.from_json(x["end"])
                if "end" in x
                else _atd_missing_json_field("MetavarValue", "end"),
                abstract_content=_atd_read_string(x["abstract_content"])
                if "abstract_content" in x
                else _atd_missing_json_field("MetavarValue", "abstract_content"),
                unique_id=UniqueId.from_json(x["unique_id"])
                if "unique_id" in x
                else _atd_missing_json_field("MetavarValue", "unique_id"),
            )
        else:
            _atd_bad_json("MetavarValue", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["start"] = (lambda x: x.to_json())(self.start)
        res["end"] = (lambda x: x.to_json())(self.end)
        res["abstract_content"] = _atd_write_string(self.abstract_content)
        res["unique_id"] = (lambda x: x.to_json())(self.unique_id)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "MetavarValue":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class MatchExtra:
    """Original type: match_extra = { ... }"""

    metavars: Dict[str, MetavarValue]
    message: Optional[str] = None

    @classmethod
    def from_json(cls, x: Any) -> "MatchExtra":
        if isinstance(x, dict):
            return cls(
                metavars=_atd_read_assoc_object_into_dict(MetavarValue.from_json)(
                    x["metavars"]
                )
                if "metavars" in x
                else _atd_missing_json_field("MatchExtra", "metavars"),
                message=_atd_read_string(x["message"]) if "message" in x else None,
            )
        else:
            _atd_bad_json("MatchExtra", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["metavars"] = _atd_write_assoc_dict_to_object(lambda x: x.to_json())(
            self.metavars
        )
        if self.message is not None:
            res["message"] = _atd_write_string(self.message)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "MatchExtra":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class Location:
    """Original type: location = { ... }"""

    path: str
    start: Position
    end: Position
    lines: List[str]

    @classmethod
    def from_json(cls, x: Any) -> "Location":
        if isinstance(x, dict):
            return cls(
                path=_atd_read_string(x["path"])
                if "path" in x
                else _atd_missing_json_field("Location", "path"),
                start=Position.from_json(x["start"])
                if "start" in x
                else _atd_missing_json_field("Location", "start"),
                end=Position.from_json(x["end"])
                if "end" in x
                else _atd_missing_json_field("Location", "end"),
                lines=_atd_read_list(_atd_read_string)(x["lines"])
                if "lines" in x
                else _atd_missing_json_field("Location", "lines"),
            )
        else:
            _atd_bad_json("Location", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["path"] = _atd_write_string(self.path)
        res["start"] = (lambda x: x.to_json())(self.start)
        res["end"] = (lambda x: x.to_json())(self.end)
        res["lines"] = _atd_write_list(_atd_write_string)(self.lines)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "Location":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class Match:
    """Original type: match_ = { ... }"""

    rule_id: RuleId
    location: Location
    extra: MatchExtra

    @classmethod
    def from_json(cls, x: Any) -> "Match":
        if isinstance(x, dict):
            return cls(
                rule_id=RuleId.from_json(x["rule_id"])
                if "rule_id" in x
                else _atd_missing_json_field("Match", "rule_id"),
                location=Location.from_json(x["location"])
                if "location" in x
                else _atd_missing_json_field("Match", "location"),
                extra=MatchExtra.from_json(x["extra"])
                if "extra" in x
                else _atd_missing_json_field("Match", "extra"),
            )
        else:
            _atd_bad_json("Match", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["rule_id"] = (lambda x: x.to_json())(self.rule_id)
        res["location"] = (lambda x: x.to_json())(self.location)
        res["extra"] = (lambda x: x.to_json())(self.extra)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "Match":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class Error:
    """Original type: error = { ... }"""

    error_type: str
    severity: Severity
    location: Location
    message: str
    rule_id: Optional[RuleId] = None
    details: Optional[str] = None
    yaml_path: Optional[List[str]] = None

    @classmethod
    def from_json(cls, x: Any) -> "Error":
        if isinstance(x, dict):
            return cls(
                error_type=_atd_read_string(x["error_type"])
                if "error_type" in x
                else _atd_missing_json_field("Error", "error_type"),
                severity=Severity.from_json(x["severity"])
                if "severity" in x
                else _atd_missing_json_field("Error", "severity"),
                location=Location.from_json(x["location"])
                if "location" in x
                else _atd_missing_json_field("Error", "location"),
                message=_atd_read_string(x["message"])
                if "message" in x
                else _atd_missing_json_field("Error", "message"),
                rule_id=RuleId.from_json(x["rule_id"]) if "rule_id" in x else None,
                details=_atd_read_string(x["details"]) if "details" in x else None,
                yaml_path=_atd_read_list(_atd_read_string)(x["yaml_path"])
                if "yaml_path" in x
                else None,
            )
        else:
            _atd_bad_json("Error", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["error_type"] = _atd_write_string(self.error_type)
        res["severity"] = (lambda x: x.to_json())(self.severity)
        res["location"] = (lambda x: x.to_json())(self.location)
        res["message"] = _atd_write_string(self.message)
        if self.rule_id is not None:
            res["rule_id"] = (lambda x: x.to_json())(self.rule_id)
        if self.details is not None:
            res["details"] = _atd_write_string(self.details)
        if self.yaml_path is not None:
            res["yaml_path"] = _atd_write_list(_atd_write_string)(self.yaml_path)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "Error":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class MatchResults:
    """Original type: match_results = { ... }"""

    matches: List[Match]
    errors: List[Error]
    skipped_targets: List[SkippedTarget]
    stats: Stats
    skipped_rules: Optional[List[SkippedRule]] = None
    time: Optional[Time] = None

    @classmethod
    def from_json(cls, x: Any) -> "MatchResults":
        if isinstance(x, dict):
            return cls(
                matches=_atd_read_list(Match.from_json)(x["matches"])
                if "matches" in x
                else _atd_missing_json_field("MatchResults", "matches"),
                errors=_atd_read_list(Error.from_json)(x["errors"])
                if "errors" in x
                else _atd_missing_json_field("MatchResults", "errors"),
                skipped_targets=_atd_read_list(SkippedTarget.from_json)(x["skipped"])
                if "skipped" in x
                else _atd_missing_json_field("MatchResults", "skipped"),
                stats=Stats.from_json(x["stats"])
                if "stats" in x
                else _atd_missing_json_field("MatchResults", "stats"),
                skipped_rules=_atd_read_list(SkippedRule.from_json)(x["skipped_rules"])
                if "skipped_rules" in x
                else None,
                time=Time.from_json(x["time"]) if "time" in x else None,
            )
        else:
            _atd_bad_json("MatchResults", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["matches"] = _atd_write_list(lambda x: x.to_json())(self.matches)
        res["errors"] = _atd_write_list(lambda x: x.to_json())(self.errors)
        res["skipped"] = _atd_write_list(lambda x: x.to_json())(self.skipped_targets)
        res["stats"] = (lambda x: x.to_json())(self.stats)
        if self.skipped_rules is not None:
            res["skipped_rules"] = _atd_write_list(lambda x: x.to_json())(
                self.skipped_rules
            )
        if self.time is not None:
            res["time"] = (lambda x: x.to_json())(self.time)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "MatchResults":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class CveResult:
    """Original type: cve_result = { ... }"""

    url: str
    filename: str
    funcnames: List[str]

    @classmethod
    def from_json(cls, x: Any) -> "CveResult":
        if isinstance(x, dict):
            return cls(
                url=_atd_read_string(x["url"])
                if "url" in x
                else _atd_missing_json_field("CveResult", "url"),
                filename=_atd_read_string(x["filename"])
                if "filename" in x
                else _atd_missing_json_field("CveResult", "filename"),
                funcnames=_atd_read_list(_atd_read_string)(x["funcnames"])
                if "funcnames" in x
                else _atd_missing_json_field("CveResult", "funcnames"),
            )
        else:
            _atd_bad_json("CveResult", x)

    def to_json(self) -> Any:
        res: Dict[str, Any] = {}
        res["url"] = _atd_write_string(self.url)
        res["filename"] = _atd_write_string(self.filename)
        res["funcnames"] = _atd_write_list(_atd_write_string)(self.funcnames)
        return res

    @classmethod
    def from_json_string(cls, x: str) -> "CveResult":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)


@dataclass(frozen=True)
class CveResults:
    """Original type: cve_results"""

    value: List[CveResult]

    @classmethod
    def from_json(cls, x: Any) -> "CveResults":
        return cls(_atd_read_list(CveResult.from_json)(x))

    def to_json(self) -> Any:
        return _atd_write_list(lambda x: x.to_json())(self.value)

    @classmethod
    def from_json_string(cls, x: str) -> "CveResults":
        return cls.from_json(json.loads(x))

    def to_json_string(self, **kw: Any) -> str:
        return json.dumps(self.to_json(), **kw)



import subprocess
from contextlib import contextmanager
from pathlib import Path
from textwrap import dedent
from textwrap import indent
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Optional

from semgrep.util import sub_check_output
from semgrep.verbose_logging import getLogger


logger = getLogger(__name__)


GIT_SH_TIMEOUT = 100


def zsplit(s: str) -> List[str]:
    """Split a string on null characters."""
    s = s.strip("\0")
    if s:
        return s.split("\0")
    else:
        return []


class GitStatus(NamedTuple):
    added: List[Path]
    modified: List[Path]
    removed: List[Path]
    unmerged: List[Path]


class StatusCode:
    Added = "A"
    Deleted = "D"
    Renamed = "R"
    Modified = "M"
    Unmerged = "U"
    TypeChanged = "T"  # changed between file / symlink / submodule
    Ignored = "!"
    Untracked = "?"
    Unstaged = " "  # but changed
    Unknown = "X"
    Broken = "B"


class BaselineHandler:
    """
    base_commit: Git ref to compare against
    """

    def __init__(self, base_commit: str) -> None:
        """
        Raises Exception if
        - cwd is not in git repo
        - base_commit is not valid git hash
        - there are tracked files with pending changes
        - there are untracked files that will be overwritten by a file in the base commit
        """
        self._base_commit = base_commit
        self._dirty_paths_by_status: Optional[Dict[str, List[Path]]] = None

        try:
            # Check commit hash exists
            try:
                subprocess.run(
                    ["git", "cat-file", "-e", base_commit],
                    check=True,
                    capture_output=True,
                )
            except subprocess.CalledProcessError:
                raise Exception(
                    dedent(
                        f"""
                        Cannot find a commit with reference '{base_commit}'. Possible reasons:

                        - the referenced commit does not exist
                        - the current working directory is not a git repository
                        - the git binary is not available

                        Try running `git show {base_commit}` to debug the issue.
                        """
                    ).strip()
                )

            self.status = self._get_git_status()
            self._abort_on_pending_changes()
            self._abort_on_conflicting_untracked_paths(self.status)
        except subprocess.CalledProcessError as e:
            raise Exception(
                f"Error initializing baseline. While running command {e.cmd} recieved non-zero exit status of {e.returncode}.\n(stdout)->{e.stdout}\n(strerr)->{e.stderr}"
            )

    def _get_git_status(self) -> GitStatus:
        """
        Read and parse git diff output to keep track of all status types
        in GitStatus object

        Paths in GitStatus object are absolute paths

        Ignores files that are symlinks to directories

        Raises CalledProcessError if there are any problems running `git diff` command
        """
        logger.debug("Initializing git status")

        # Output of git command will be relative to git project root not cwd
        logger.debug("Running git diff")
        status_cmd = [
            "git",
            "diff",
            "--cached",
            "--name-status",
            "--no-ext-diff",
            "-z",
            "--diff-filter=ACDMRTUXB",
            "--ignore-submodules",
            "--relative",
            self._base_commit,
        ]
        try:
            # nosemgrep: python.lang.security.audit.dangerous-subprocess-use.dangerous-subprocess-use
            raw_output = subprocess.run(
                [*status_cmd, "--merge-base"],
                timeout=GIT_SH_TIMEOUT,
                capture_output=True,
                encoding="utf-8",
                check=True,
            ).stdout

        except subprocess.CalledProcessError as exc:
            if exc.stderr.strip() == "fatal: multiple merge bases found":
                logger.warn(
                    "git could not find a single branch-off point, so we will compare the baseline commit directly"
                )
                # nosemgrep: python.lang.security.audit.dangerous-subprocess-use.dangerous-subprocess-use
                raw_output = subprocess.run(
                    status_cmd,
                    timeout=GIT_SH_TIMEOUT,
                    capture_output=True,
                    encoding="utf-8",
                    check=True,
                ).stdout
            else:
                raise exc
        status_output = zsplit(raw_output)
        logger.debug("Finished git diff. Parsing git status output")
        logger.debug(status_output)
        added = []
        modified = []
        removed = []
        unmerged = []
        while status_output:
            code = status_output.pop(0)
            fname = status_output.pop(0)
            new_fname = None

            # code is RXXX, where XXX is percent similarity
            if code[0] == StatusCode.Renamed:
                new_fname = status_output.pop(0)

            if not code.strip():
                continue
            if code == StatusCode.Untracked or code == StatusCode.Ignored:
                continue

            path = Path(fname)

            if path.is_symlink() and path.is_dir():
                logger.verbose(
                    f"| Skipping {path} since it is a symlink to a directory: {path.resolve()}",
                )
                continue
            # The following detection for unmerged codes comes from `man git-status`
            if code == StatusCode.Unmerged:
                unmerged.append(path)
            # code is RXXX, where XXX is percent similarity
            if code[0] == StatusCode.Renamed and new_fname:
                removed.append(path)
                added.append(Path(new_fname))
            if code == StatusCode.Added:
                added.append(path)
            if code == StatusCode.Modified:
                modified.append(path)
            if code == StatusCode.TypeChanged and not path.is_symlink():
                modified.append(path)
            if code == StatusCode.Deleted:
                removed.append(path)

        logger.debug(
            f"Git status:\nadded: {added}\nmodified: {modified}\nremoved: {removed}\nunmerged: {unmerged}"
        )

        return GitStatus(added, modified, removed, unmerged)

    def _get_dirty_paths_by_status(self) -> Dict[str, List[Path]]:
        """
        Returns all paths that have a git status, grouped by change type.

        Raises CalledProcessError if `git status` command fails though
        not clear how that would happen since at this point cwd must be valid repo

        These can be staged, unstaged, or untracked.
        """
        if self._dirty_paths_by_status is not None:
            return self._dirty_paths_by_status

        logger.debug("Initializing dirty paths")
        sub_out = subprocess.run(
            ["git", "status", "--porcelain", "-z"],
            timeout=GIT_SH_TIMEOUT,
            capture_output=True,
            encoding="utf-8",
            check=True,
        )
        git_status_output = sub_out.stdout
        logger.debug(f"Git status output: {git_status_output}")
        output = zsplit(git_status_output)
        logger.debug("finished getting dirty paths")

        dirty_paths: Dict[str, List[Path]] = {}
        for line in output:
            status_code = line[0]
            path = Path(line[3:])

            if status_code in dirty_paths:
                dirty_paths[status_code].append(path)
            else:
                dirty_paths[status_code] = [path]

        logger.debug(str(dirty_paths))

        # Cache dirty paths
        self._dirty_paths_by_status = dirty_paths
        return dirty_paths

    def _abort_on_pending_changes(self) -> None:
        """
        Raises Exception if any tracked files are changed.

        We are aborting for now to prevent inadvertently deleting files while
        doing a baseline scan until we confidently stash and unstash said changes
        """
        if set(self._get_dirty_paths_by_status()) - {StatusCode.Untracked}:
            raise Exception(
                "Found pending changes in tracked files. Baseline scans runs require a clean git state."
            )

    def _abort_on_conflicting_untracked_paths(self, status: GitStatus) -> None:
        """
        Raises Exception if untracked paths in head were touched in baseline commit
        This would mean checking out the baseline would overwrite said file

        :raises Exception: If the git repo is not in a clean state
        """
        changed_paths = set(
            status.added + status.modified + status.removed + status.unmerged
        )
        untracked_paths = {
            str(path)
            for path in (
                self._get_dirty_paths_by_status().get(StatusCode.Untracked, [])
            )
        }
        overlapping_paths = untracked_paths & changed_paths

        if overlapping_paths:
            raise Exception(
                f"Found files that are untracked by git but exist in {self._base_commit}",
                "Running a baseline scan will cause changes to be overwritten, so aborting.",
                f"Please commit or stash your untracked changes in these paths: {overlapping_paths}.",
            )

    @contextmanager
    def baseline_context(self) -> Iterator[None]:
        """
        Yields context where pwd is modified to be self.commit_hash
        upon exiting the context returns pwd to what it was initially

        Usage:

        bh = BaselineHandler(commit_hash)
        with baseline_context():
            # Do stuff here
            # pwd will be on commit_hash
        # pwd will be back to what it was


        Raises CalledProcessError if any calls to git return non-zero exit code
        """
        # Reabort in case for some reason aborting in __init__ did not cause
        # semgrep to exit
        self._abort_on_pending_changes()
        self._abort_on_conflicting_untracked_paths(self.status)

        logger.debug("Running git write-tree")
        current_head = subprocess.run(
            ["git", "rev-parse", "HEAD"],
            timeout=GIT_SH_TIMEOUT,
            capture_output=True,
            encoding="utf-8",
            check=True,
        ).stdout.strip()
        try:
            merge_base_sha = (
                sub_check_output(["git", "merge-base", self._base_commit, "HEAD"])
                .rstrip()
                .decode()
            )

            logger.debug("Running git checkout for baseline context")
            subprocess.run(
                ["git", "reset", "--hard", merge_base_sha],
                timeout=GIT_SH_TIMEOUT,
                capture_output=True,
                check=True,
            )
            logger.debug("Finished git checkout for baseline context")
            yield
        finally:
            # Return to non-baseline state

            # git checkout will fail if the checked-out index deletes all files in the repo
            # In this case, we still want to continue without error.
            # Note that we have no good way of detecting this issue without inspecting the checkout output
            # message, which means we are fragile with respect to git version here.
            logger.debug("Running git reset to return original context")
            x = subprocess.run(
                ["git", "reset", "--hard", current_head],
                capture_output=True,
                timeout=GIT_SH_TIMEOUT,
            )
            logger.debug("Finished git reset to return original context")

            if x.returncode != 0:
                output = x.stderr.decode()
                if (
                    output
                    and len(output) >= 2
                    and "pathspec '.' did not match any file(s) known to git"
                    in output.strip()
                ):
                    logger.debug(
                        "Restoring git index failed due to total repository deletion; skipping checkout"
                    )
                else:
                    raise Exception(
                        f"Fatal error restoring Git state; please restore your repository state manually:\n{output}"
                    )

    def print_git_log(self) -> None:
        base_commit_sha = (
            sub_check_output(["git", "rev-parse", self._base_commit]).rstrip().decode()
        )
        merge_base_sha = (
            sub_check_output(["git", "merge-base", self._base_commit, "HEAD"])
            .rstrip()
            .decode()
        )
        logger.info("  Will report findings introduced by these commits:")
        log = sub_check_output(
            ["git", "log", "--oneline", "--graph", f"{merge_base_sha}..HEAD"],
            timeout=GIT_SH_TIMEOUT,
            encoding="utf-8",
        ).rstrip()
        logger.info(indent(log, "    "))
        if merge_base_sha != base_commit_sha:
            logger.warning(
                "  The current branch is missing these commits from the baseline branch:"
            )
            log = sub_check_output(
                [
                    "git",
                    "log",
                    "--oneline",
                    "--graph",
                    f"{merge_base_sha}..{base_commit_sha}",
                ],
                timeout=GIT_SH_TIMEOUT,
                encoding="utf-8",
            )
            logger.info(indent(log, "    ").rstrip())

            logger.info(
                "  Any finding these commits fixed will look like a new finding in the current branch."
            )
            logger.info(
                "  To avoid reporting such findings, compare to the branch-off point with:\n"
                f"    --baseline-commit=$(git merge-base {self._base_commit} HEAD)"
            )



from typing import Any
from typing import Mapping

from semgrep.formatter.gitlab_sast import GitlabSastFormatter
from semgrep.rule_match import RuleMatch


class GitlabSecretsFormatter(GitlabSastFormatter):
    def _format_rule_match(self, rule_match: RuleMatch) -> Mapping[str, Any]:
        return {
            **super()._format_rule_match(rule_match),
            "category": "secret_detection",
            "raw_source_code_extract": rule_match.lines,
            "commit": {
                "date": "1970-01-01T00:00:00Z",
                # Even the native Gitleaks based Gitlab secret detection
                # only provides a dummy value for now on relevant hash.
                "sha": "0000000",
            },
        }



import functools
import itertools
import logging
import operator
import os
import subprocess
import sys
from pathlib import Path
from typing import Any
from typing import Callable
from typing import FrozenSet
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import TypeVar
from urllib.parse import urlparse

import click

from semgrep.constants import Colors
from semgrep.constants import USER_LOG_FILE
from semgrep.constants import YML_SUFFIXES
from semgrep.constants import YML_TEST_SUFFIXES

T = TypeVar("T")

global FORCE_COLOR
FORCE_COLOR = False
global FORCE_NO_COLOR
FORCE_NO_COLOR = False

global VERBOSITY
VERBOSITY = logging.INFO


MAX_TEXT_WIDTH = 120


def is_quiet() -> bool:
    """
    Returns true if logging level is quiet or quieter (higher)
    (i.e. only critical logs surfaced)
    """
    return VERBOSITY >= logging.CRITICAL


def is_debug() -> bool:
    """
    Returns true if logging level is debug or noisier (lower)
    (i.e. want more logs)
    """
    return VERBOSITY <= logging.DEBUG


def is_url(url: str) -> bool:
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def set_flags(*, verbose: bool, debug: bool, quiet: bool, force_color: bool) -> None:
    """Set the relevant logging levels"""
    # Assumes only one of verbose, debug, quiet is True
    logger = logging.getLogger("semgrep")
    logger.handlers = []  # Reset to no handlers

    stdout_level = logging.INFO
    if verbose:
        stdout_level = logging.VERBOSE  # type: ignore[attr-defined]
    elif debug:
        stdout_level = logging.DEBUG
    elif quiet:
        stdout_level = logging.CRITICAL

    # Setup stdout logging
    stdout_handler = logging.StreamHandler()
    stdout_formatter = logging.Formatter("%(message)s")
    stdout_handler.setFormatter(stdout_formatter)
    stdout_handler.setLevel(stdout_level)
    logger.addHandler(stdout_handler)

    # Setup file logging
    # USER_LOG_FILE dir must exist
    USER_LOG_FILE.parent.mkdir(parents=True, exist_ok=True)
    file_handler = logging.FileHandler(USER_LOG_FILE, "w")
    file_formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(file_formatter)
    logger.addHandler(file_handler)

    # Needs to be DEBUG otherwise will filter before sending to handlers
    logger.setLevel(logging.DEBUG)

    global VERBOSITY
    VERBOSITY = stdout_level

    global FORCE_COLOR
    if force_color or os.environ.get("SEMGREP_FORCE_COLOR") is not None:
        FORCE_COLOR = True

    global FORCE_NO_COLOR
    if (
        os.environ.get("NO_COLOR") is not None  # https://no-color.org/
        or os.environ.get("SEMGREP_FORCE_NO_COLOR") is not None
    ):
        FORCE_NO_COLOR = True


def partition(
    pred: Callable[[T], Any], iterable: Iterable[T]
) -> Tuple[List[T], List[T]]:
    """E.g. partition(is_odd, range(10)) -> 1 3 5 7 9  and  0 2 4 6 8"""
    i1, i2 = itertools.tee(iterable)
    return list(filter(pred, i1)), list(itertools.filterfalse(pred, i2))


def partition_set(
    pred: Callable[[T], Any], iterable: Iterable[T]
) -> Tuple[FrozenSet[T], FrozenSet[T]]:
    """E.g. partition(is_odd, range(10)) -> 1 3 5 7 9  and  0 2 4 6 8"""
    i1, i2 = itertools.tee(iterable)
    return frozenset(filter(pred, i1)), frozenset(itertools.filterfalse(pred, i2))


def abort(message: str) -> None:
    click.secho(message, fg="red", err=True)
    sys.exit(2)


def with_color(
    color: Colors,
    text: str,
    bgcolor: Optional[Colors] = None,
    bold: bool = False,
    underline: bool = False,
) -> str:
    """
    Wrap text in color & reset

    Use ANSI color names or 8 bit colors (24-bit is not well supported by terminals)
    In click bold always switches colors to their bright variant (if there is one)
    """
    if FORCE_NO_COLOR and not FORCE_COLOR:
        # for 'no color' there is a global env var (https://no-color.org/)
        # while the env var to 'force color' is semgrep-specific
        # so we let the more specific setting override the broader one
        return text
    if not sys.stderr.isatty() and not FORCE_COLOR:
        return text
    return click.style(
        text,
        fg=color.value,
        bg=(bgcolor.value if bgcolor is not None else None),
        underline=underline,
        bold=bold,
    )


def terminal_wrap(text: str) -> str:
    from shutil import get_terminal_size
    import textwrap

    paras = text.split("\n")
    terminal_size = get_terminal_size((MAX_TEXT_WIDTH, 1))[0]
    if terminal_size <= 0:
        terminal_size = MAX_TEXT_WIDTH
    width = min(MAX_TEXT_WIDTH, terminal_size)
    wrapped_paras = ["\n".join(textwrap.wrap(p, width)) for p in paras]
    return "\n".join(wrapped_paras)


def sub_check_output(cmd: List[str], **kwargs: Any) -> Any:
    """A simple proxy function to minimize and centralize subprocess usage."""
    if is_quiet():
        kwargs = {**kwargs, "stderr": subprocess.DEVNULL}

    # nosemgrep: python.lang.security.audit.dangerous-subprocess-use.dangerous-subprocess-use
    return subprocess.check_output(cmd, **kwargs)


def manually_search_file(path: str, search_term: str, suffix: str) -> Optional[str]:
    """
    Searches a file for the given search term and, if found,
    returns the first word that contains that search term
    """
    if not os.path.isfile(path):
        return None
    with open(path) as fd:
        contents = fd.read()
        words = contents.split()
    # Find all of the individual words that contain the search_term
    matches = [w for w in words if search_term in w]
    return matches[0] + suffix if len(matches) > 0 else None


# TODO: seems dead
def listendswith(l: List[T], tail: List[T]) -> bool:
    """
    E.g.
        - listendswith([1, 2, 3, 4], [3, 4]) -> True
        - listendswith([1, 2, 3, 4], [1, 4]) -> False
    """
    if len(tail) > len(l):
        return False

    return all(l[len(l) - len(tail) + i] == tail[i] for i in range(len(tail)))


def is_config_suffix(path: Path) -> bool:
    return any(
        listendswith(path.suffixes, suffixes) for suffixes in YML_SUFFIXES
    ) and not is_config_test_suffix(path)


def is_config_test_suffix(path: Path) -> bool:
    return any(listendswith(path.suffixes, suffixes) for suffixes in YML_TEST_SUFFIXES)


def format_bytes(num: float) -> str:
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1024.0:
            return "%3d%sB" % (num, unit)
        num /= 1024.0
    return "{:.1f}{}B".format(num, "Y")


def truncate(file_name: str, col_lim: int) -> str:
    name_len = len(file_name)
    prefix = "..."
    if name_len > col_lim:
        file_name = prefix + file_name[name_len - col_lim + len(prefix) :]
    return file_name


def flatten(some_list: List[List[T]]) -> List[T]:
    return functools.reduce(operator.iconcat, some_list, [])


PathFilterCallable = Callable[..., FrozenSet[Path]]


def unit_str(count: int, unit: str, pad: bool = False) -> str:
    if count != 1:
        unit += "s"
    elif pad:
        unit += " "

    return f"{count} {unit}"



from typing import Dict
from typing import List

import click
import requests

from semgrep.constants import RETURNTOCORP_LEVER_URL


def print_job_postings() -> None:
    # http get postings
    response = requests.get(RETURNTOCORP_LEVER_URL, timeout=30)
    postings = response.json()

    # organise per team
    postings_dict: Dict[str, List[Dict[str, str]]] = {}
    for posting in postings:
        team = posting["categories"]["team"]
        if team not in postings_dict:
            postings_dict[team] = []

        title = posting["text"]
        url = posting["hostedUrl"]
        postings_dict[team].append({"title": title, "url": url})

    click.echo(
        "We love that you're using Semgrep. Would you like to... work on it full time?! 😄 ",
        err=True,
    )
    click.echo("", err=True)
    # iterate over teams alphabetically
    for team in sorted(postings_dict):
        team_character_count = len(team)

        click.echo((team_character_count + 4) * "#", err=True)
        click.echo("# " + team + " #", err=True)
        click.echo((team_character_count + 4) * "#", err=True)
        click.echo("", err=True)

        for posting in postings_dict[team]:
            click.echo("* " + posting["title"] + ": " + posting["url"], err=True)

        click.echo("", err=True)

    click.echo("", err=True)
    click.echo(
        "PS: Just wanted to apply autofixes? Use --autofix instead of --apply", err=True
    )



import contextlib
import os
import subprocess
import sys
import tempfile
from collections import defaultdict
from functools import lru_cache
from functools import partial
from pathlib import Path
from typing import Any
from typing import Callable
from typing import cast
from typing import Collection
from typing import Dict
from typing import FrozenSet
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Mapping
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple

from semgrep.git import BaselineHandler

# usually this would be a try...except ImportError
# but mypy understands only this
# see https://github.com/python/mypy/issues/1393
if sys.version_info[:2] >= (3, 8):
    # Literal is available in stdlib since Python 3.8
    from typing import Literal
else:
    from typing_extensions import Literal

from attrs import define
from attrs import field
import click
from attrs import Factory, frozen
from wcmatch import glob as wcglob

from semgrep.constants import Colors
from semgrep.error import FilesNotFoundError
from semgrep.formatter.text import width
from semgrep.ignores import FileIgnore
from semgrep.semgrep_types import FileExtension
from semgrep.semgrep_types import LANGUAGE
from semgrep.semgrep_types import Language
from semgrep.semgrep_types import Shebang
from semgrep.types import FilteredFiles
from semgrep.util import partition_set
from semgrep.util import sub_check_output
from semgrep.util import with_color
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)

MAX_CHARS_TO_READ_FOR_SHEBANG = 255
PATHS_ALWAYS_SKIPPED = (".git",)

ALL_EXTENSIONS: Collection[FileExtension] = {
    ext
    for definition in LANGUAGE.definition_by_id.values()
    for ext in definition.exts
    if ext != FileExtension("")
}


@contextlib.contextmanager
def converted_pipe_targets(targets: Sequence[str]) -> Iterator[Sequence[str]]:
    """
    Provides a context in which FIFOs have been copied into temp files

    This is necessary as we can not easily rewire these pipes into the called semgrep-core
    process.

    :param targets: Input target specifiers
    :return: A sequence of non-pipe specifiers (Path(t).is_file() returns true)
    """

    out_targets = []
    with tempfile.TemporaryDirectory() as temp_dir:
        for t in targets:
            if t == "-":
                with (Path(temp_dir) / "stdin").open("wb") as fd:
                    fd.write(sys.stdin.buffer.read())
                out_targets.append(fd.name)
            elif Path(t).is_fifo():
                with (Path(temp_dir) / t[1:].replace("/", "_")).open("wb") as fd:
                    with Path(t).open("rb") as td:
                        fd.write(td.read())
                out_targets.append(fd.name)
            else:
                out_targets.append(t)
        yield out_targets


@define
class FileTargetingLog:
    """Keeps track of which paths were ignored for what reason.

    Each attribute is a distinct reason why files could be ignored.

    Some reason can apply once per rule; these are mappings keyed on the rule id.
    """

    target_manager: "TargetManager"

    semgrepignored: Set[Path] = Factory(set)
    always_skipped: Set[Path] = Factory(set)
    cli_includes: Set[Path] = Factory(set)
    cli_excludes: Set[Path] = Factory(set)
    size_limit: Set[Path] = Factory(set)
    failed_to_analyze: Set[Path] = Factory(set)

    by_language: Dict[Language, Set[Path]] = Factory(lambda: defaultdict(set))
    rule_includes: Dict[str, Set[Path]] = Factory(lambda: defaultdict(set))
    rule_excludes: Dict[str, Set[Path]] = Factory(lambda: defaultdict(set))

    @property
    def rule_ids_with_skipped_paths(self) -> FrozenSet[str]:
        """All rule IDs that have skipped paths.

        Note that if a rule ID defines excludes/includes,
        but they didn't skip any paths,
        that rule ID will not show up here.
        """
        return frozenset(
            {
                *(rule_id for rule_id, skips in self.rule_includes.items() if skips),
                *(rule_id for rule_id, skips in self.rule_excludes.items() if skips),
            }
        )

    def __str__(self) -> str:
        limited_fragments = []
        skip_fragments = []

        if self.target_manager.baseline_handler:
            limited_fragments.append(
                "Scan was limited to files changed since baseline commit."
            )
        elif self.target_manager.respect_git_ignore:
            limited_fragments.append("Scan was limited to files tracked by git.")

        if self.cli_includes:
            skip_fragments.append(
                f"{len(self.cli_includes)} files not matching --include patterns"
            )
        if self.cli_excludes:
            skip_fragments.append(
                f"{len(self.cli_excludes)} files matching --exclude patterns"
            )
        if self.size_limit:
            skip_fragments.append(
                f"{len(self.size_limit)} files larger than {self.target_manager.max_target_bytes / 1000 / 1000} MB"
            )
        if self.semgrepignored:
            skip_fragments.append(
                f"{len(self.semgrepignored)} files matching .semgrepignore patterns"
            )
        if self.failed_to_analyze:
            skip_fragments.append(
                f"{len(self.failed_to_analyze)} files not analyzed due to a parsing or internal Semgrep error"
            )

        if not limited_fragments and not skip_fragments:
            return ""

        message = "Some files were skipped."
        if limited_fragments:
            for fragment in limited_fragments:
                message += f"\n  {fragment}"
        if skip_fragments:
            message += "\n  Scan skipped: " + ", ".join(skip_fragments)
            message += "\n  For a full list of skipped files, run semgrep with the --verbose flag."
        message += "\n"
        return message

    def yield_verbose_lines(self) -> Iterator[Tuple[Literal[0, 1, 2], str]]:
        """Yields lines of verbose output for the skipped files.

        The returned tuple is (level, message).
        The level is a number; one of 0, 1, or 2, which sets the indentation when outputting the line.
        """
        yield 0, "Files skipped:"

        yield 1, "Always skipped by Semgrep:"
        if self.always_skipped:
            for path in self.always_skipped:
                yield 2, with_color(Colors.cyan, str(path))
        else:
            yield 2, "<none>"

        yield 1, "Skipped by .gitignore:"
        if self.target_manager.respect_git_ignore:
            yield 1, "(Disable by passing --no-git-ignore)"
            yield 2, "<all files not listed by `git ls-files` were skipped>"
        else:
            yield 1, "(Disabled with --no-git-ignore)"
            yield 2, "<none>"

        yield 1, "Skipped by .semgrepignore:"
        yield 1, "(See: https://semgrep.dev/docs/ignoring-files-folders-code/#understanding-semgrep-defaults)"
        if self.semgrepignored:
            for path in self.semgrepignored:
                yield 2, with_color(Colors.cyan, str(path))
        else:
            yield 2, "<none>"

        yield 1, "Skipped by --include patterns:"
        if self.cli_includes:
            for path in self.cli_includes:
                yield 2, with_color(Colors.cyan, str(path))
        else:
            yield 2, "<none>"

        yield 1, "Skipped by --exclude patterns:"
        if self.cli_excludes:
            for path in self.cli_excludes:
                yield 2, with_color(Colors.cyan, str(path))
        else:
            yield 2, "<none>"

        yield 1, f"Skipped by limiting to files smaller than {self.target_manager.max_target_bytes} bytes:"
        yield 1, "(Adjust with the --max-target-bytes flag)"
        if self.size_limit:
            for path in self.size_limit:
                yield 2, with_color(Colors.cyan, str(path))
        else:
            yield 2, "<none>"

        yield 1, "Skipped by analysis failure due to parsing or internal Semgrep error"
        if self.failed_to_analyze:
            for path in self.failed_to_analyze:
                yield 2, with_color(Colors.cyan, str(path))
        else:
            yield 2, "<none>"

    def verbose_output(self) -> str:
        formatters_by_level: Mapping[int, Callable[[str], str]] = {
            0: lambda line: "\n".join([40 * "=", line, 40 * "="]),
            1: lambda line: click.wrap_text(
                with_color(Colors.foreground, line, bold=True),
                width,
                2 * " ",
                2 * " ",
                False,
            ),
            2: lambda line: click.wrap_text(
                line,
                width,
                "   • ",
                "     ",
                False,
            ),
        }
        output = ""

        prev_level = None
        for level, line in self.yield_verbose_lines():
            if prev_level != level:
                output += "\n"
            formatter = formatters_by_level[level]
            output += formatter(line) + "\n"
            prev_level = level

        return output

    def yield_json_objects(self) -> Iterable[Dict[str, Any]]:
        for path in self.always_skipped:
            yield {"path": str(path), "reason": "always_skipped"}
        for path in self.semgrepignored:
            yield {"path": str(path), "reason": "semgrepignore_patterns_match"}
        for path in self.cli_includes:
            yield {"path": str(path), "reason": "cli_include_flags_do_not_match"}
        for path in self.cli_excludes:
            yield {"path": str(path), "reason": "cli_exclude_flags_match"}
        for path in self.size_limit:
            yield {
                "path": str(path),
                "reason": "exceeded_size_limit",
                "size_limit_bytes": self.target_manager.max_target_bytes,
            }
        for path in self.failed_to_analyze:
            yield {
                "path": str(path),
                "reason": "analysis_failed_parser_or_internal_error",
            }


@frozen(eq=False)  #
class Target:
    """
    Represents one path that was given as a target.
    Then target.paths returns all paths that target expands to.
    This does not do any include/exclude filtering.

    Three strategies are available for gathering targets:
    1. recursively collect from file system (slowest, but always works)
    2. read the output of `git ls-files` (respects .gitignore)
    3. [TODO] read the output of `git diff` (respects --baseline-commit)
    """

    path: Path = field(converter=Path)
    git_tracked_only: bool = False
    baseline_handler: Optional[BaselineHandler] = None

    @path.validator
    def validate_path(self, _: Any, value: Path) -> None:
        """
        Check whether the targeted path exists.

        If not, the path might be a socket.
        """
        if not self._is_valid_file_or_dir(value):
            raise FilesNotFoundError(paths=tuple([value]))
        return None

    def _is_valid_file_or_dir(self, path: Path) -> bool:
        """Check this is a valid file or directory for semgrep scanning."""
        return os.access(str(path), os.R_OK) and not path.is_symlink()

    def _is_valid_file(self, path: Path) -> bool:
        """Check if file is a readable regular file.

        This eliminates files that should never be semgrep targets. Among
        others, this takes care of excluding symbolic links (because we don't
        want to scan the target twice), directories (which may be returned by
        globbing or by 'git ls-files' e.g. submodules), and files missing
        the read permission.
        """
        return self._is_valid_file_or_dir(path) and path.is_file()

    def _parse_git_output(self, output: str) -> FrozenSet[Path]:
        """
        Convert a newline delimited list of files to a set of path objects
        prepends curr_dir to all paths in said list

        If list is empty then returns an empty set
        """
        files: FrozenSet[Path] = frozenset()
        if output:
            files = frozenset(
                p
                for p in (self.path / elem for elem in output.strip().split("\n"))
                if self._is_valid_file(p)
            )
        return files

    def files_from_git_diff(self) -> FrozenSet[Path]:
        """
        Get only changed files since baseline commit.
        """
        if self.baseline_handler is None:
            raise RuntimeError("Can't get git diff file list without a baseline commit")
        git_status = self.baseline_handler.status
        return frozenset(git_status.added + git_status.modified)

    def files_from_git_ls(self) -> FrozenSet[Path]:
        """
        git ls-files is significantly faster than os.walk when performed on a git project,
        so identify the git files first, then filter those
        """
        run_git_command = partial(
            sub_check_output,
            cwd=self.path.resolve(),
            encoding="utf-8",
            stderr=subprocess.DEVNULL,
        )

        # Tracked files
        tracked_output = run_git_command(["git", "ls-files"])

        # Untracked but not ignored files
        untracked_output = run_git_command(
            [
                "git",
                "ls-files",
                "--other",
                "--exclude-standard",
            ]
        )

        deleted_output = run_git_command(["git", "ls-files", "--deleted"])
        tracked = self._parse_git_output(tracked_output)
        untracked_unignored = self._parse_git_output(untracked_output)
        deleted = self._parse_git_output(deleted_output)
        return frozenset(tracked | untracked_unignored - deleted)

    def files_from_filesystem(self) -> FrozenSet[Path]:
        return frozenset(
            match
            for match in self.path.glob("**/*")
            if match.is_file() and not match.is_symlink()
        )

    @lru_cache(maxsize=None)
    def files(self) -> FrozenSet[Path]:
        """
        Recursively go through a directory and return list of all files with
        default file extension of language
        """
        if not self.path.is_dir() and self.path.is_file():
            return frozenset([self.path])

        if self.baseline_handler is not None:
            try:
                return self.files_from_git_diff()
            except (subprocess.CalledProcessError, FileNotFoundError):
                logger.verbose(
                    f"Unable to target only the changed files since baseline commit. Running on all git tracked files instead..."
                )

        if self.git_tracked_only:
            try:
                return self.files_from_git_ls()
            except (subprocess.CalledProcessError, FileNotFoundError):
                logger.verbose(
                    f"Unable to ignore files ignored by git ({self.path} is not a git directory or git is not installed). Running on all files instead..."
                )

        return self.files_from_filesystem()


@define(eq=False)
class TargetManager:
    """
    Handles all file include/exclude logic for semgrep

    Assumes file system does not change during it's existence to cache
    files for a given language etc. If file system changes (i.e. git checkout),
    create a new TargetManager object

    If respect_git_ignore is true then will only consider files that are
    tracked or (untracked but not ignored) by git

    If git_baseline_commit is true then will only consider files that have
    changed since that commit

    If allow_unknown_extensions is set then targets with extensions that are
    not understood by semgrep will always be returned by get_files. Else will discard
    targets with unknown extensions

    TargetManager not to be confused with https://jobs.target.com/search-jobs/store%20manager
    """

    target_strings: Sequence[str]
    includes: Sequence[str] = Factory(list)
    excludes: Sequence[str] = Factory(list)
    max_target_bytes: int = -1
    respect_git_ignore: bool = False
    baseline_handler: Optional[BaselineHandler] = None
    allow_unknown_extensions: bool = False
    file_ignore: Optional[FileIgnore] = None
    ignore_log: FileTargetingLog = Factory(FileTargetingLog, takes_self=True)
    targets: Sequence[Target] = field(init=False)

    _filtered_targets: Dict[Language, FilteredFiles] = field(factory=dict)

    def __attrs_post_init__(self) -> None:
        self.targets = [
            Target(
                target,
                git_tracked_only=self.respect_git_ignore,
                baseline_handler=self.baseline_handler,
            )
            for target in self.target_strings
        ]
        return None

    @staticmethod
    def preprocess_path_patterns(patterns: Sequence[str]) -> List[str]:
        """Convert semgrep's path include/exclude patterns to wcmatch's glob patterns.

        In semgrep, pattern "foo/bar" should match paths "x/foo/bar", "foo/bar/x", and
        "x/foo/bar/x". It implicitly matches zero or more directories at the beginning and the end
        of the pattern. In contrast, we have to explicitly specify the globstar (**) patterns in
        wcmatch. This function will converts a pattern "foo/bar" into "**/foo/bar" and
        "**/foo/bar/**". We need the pattern without the trailing "/**" because "foo/bar.py/**"
        won't match "foo/bar.py".
        """
        result = []
        for pattern in patterns:
            result.append("**/" + pattern)
            result.append("**/" + pattern + "/**")
        return result

    @staticmethod
    def _executes_with_shebang(f: Path, shebangs: Collection[Shebang]) -> bool:
        """
        Returns if a path is executable and executes with one of a set of programs
        """
        if not os.access(str(f), os.X_OK | os.R_OK):
            return False
        try:
            with f.open("r") as fd:
                hline = fd.readline(MAX_CHARS_TO_READ_FOR_SHEBANG).rstrip()
            return any(hline.endswith(s) for s in shebangs)
        except UnicodeDecodeError:
            logger.debug(
                f"Encountered likely binary file {f} while reading shebang; skipping this file"
            )
            return False

    def filter_by_language(
        self, language: Language, *, candidates: FrozenSet[Path]
    ) -> FilteredFiles:
        """
        Returns only paths that have the correct extension or shebang.

        Finds all files in a collection of paths that either:
        - end with one of a set of extension
        - is a script that executes with one of a set of programs
        """
        kept = frozenset(
            path
            for path in candidates
            if any(str(path).endswith(ext) for ext in language.definition.exts)
            or self._executes_with_shebang(path, language.definition.shebangs)
        )
        return FilteredFiles(kept, frozenset(candidates - kept))

    def filter_known_extensions(self, *, candidates: FrozenSet[Path]) -> FilteredFiles:
        """
        Returns only paths that have an extension we don't recognize.
        """
        kept = frozenset(
            path
            for path in candidates
            if not any(path.match(f"*{ext}") for ext in ALL_EXTENSIONS)
        )
        return FilteredFiles(kept, frozenset(candidates - kept))

    @staticmethod
    def filter_includes(
        includes: Sequence[str], *, candidates: FrozenSet[Path]
    ) -> FilteredFiles:
        """
        Returns all elements in candidates that match any includes pattern

        If includes is empty, returns candidates unchanged
        """
        if not includes:
            return FilteredFiles(candidates)

        includes = TargetManager.preprocess_path_patterns(includes)
        # Need cast b/c types-wcmatch doesn't use generics properly :(
        kept = frozenset(
            cast(
                Iterable[Path],
                wcglob.globfilter(
                    candidates, includes, flags=wcglob.GLOBSTAR | wcglob.DOTGLOB
                ),
            )
        )
        return FilteredFiles(kept, frozenset(candidates - kept))

    @staticmethod
    def filter_excludes(
        excludes: Sequence[str], *, candidates: FrozenSet[Path]
    ) -> FilteredFiles:
        """
        Returns all elements in candidates that do not match any excludes pattern

        If excludes is empty, returns candidates unchanged
        """
        if not excludes:
            return FilteredFiles(candidates)

        excludes = TargetManager.preprocess_path_patterns(excludes)

        # Need cast b/c types-wcmatch doesn't use generics properly :(
        removed = frozenset(
            cast(
                Iterable[Path],
                wcglob.globfilter(
                    candidates, excludes, flags=wcglob.GLOBSTAR | wcglob.DOTGLOB
                ),
            )
        )
        return FilteredFiles(frozenset(candidates - removed), removed)

    @staticmethod
    def filter_by_size(
        max_target_bytes: int, *, candidates: FrozenSet[Path]
    ) -> FilteredFiles:
        """
        Return all the files whose size doesn't exceed the limit.

        If max_target_bytes is zero or negative, all paths are returned.
        If some paths are invalid, they may or may not be included in the
        result.
        """
        if max_target_bytes <= 0:
            return FilteredFiles(candidates)

        kept, removed = partition_set(
            lambda path: os.path.getsize(path) <= max_target_bytes,
            candidates,
        )

        return FilteredFiles(kept, removed)

    @lru_cache(maxsize=None)
    def get_files_for_language(self, lang: Language) -> FilteredFiles:
        """
        Return all files that are decendants of any directory in TARGET that have
        an extension matching LANG that match any pattern in INCLUDES and do not
        match any pattern in EXCLUDES. Any file in TARGET bypasses excludes and includes.
        If a file in TARGET has a known extension that is not for langugage LANG then
        it is also filtered out

        Note also filters out any directory and descendants of `.git`
        """
        all_files = frozenset(f for target in self.targets for f in target.files())

        files = self.filter_by_language(lang, candidates=all_files)
        self.ignore_log.by_language[lang].update(files.removed)

        files = self.filter_includes(self.includes, candidates=files.kept)
        self.ignore_log.cli_includes.update(files.removed)

        files = self.filter_excludes(self.excludes, candidates=files.kept)
        self.ignore_log.cli_excludes.update(files.removed)

        files = self.filter_excludes(PATHS_ALWAYS_SKIPPED, candidates=files.kept)
        self.ignore_log.always_skipped.update(files.removed)

        files = self.filter_by_size(self.max_target_bytes, candidates=files.kept)
        self.ignore_log.size_limit.update(files.removed)

        if self.file_ignore:
            files = self.file_ignore.filter_paths(candidates=files.kept)
            self.ignore_log.semgrepignored.update(files.removed)

        kept_files = files.kept

        explicit_files = frozenset(
            t.path for t in self.targets if not t.path.is_dir() and t.path.is_file()
        )
        explicit_files_for_lang = self.filter_by_language(
            lang, candidates=explicit_files
        )
        kept_files |= explicit_files_for_lang.kept
        if self.allow_unknown_extensions:
            explicit_files_of_unknown_lang = self.filter_known_extensions(
                candidates=explicit_files
            )
            kept_files |= explicit_files_of_unknown_lang.kept

        return FilteredFiles(kept_files, all_files - kept_files)

    def get_files_for_rule(
        self,
        lang: Language,
        rule_includes: Sequence[str],
        rule_excludes: Sequence[str],
        rule_id: str,
    ) -> FrozenSet[Path]:
        """
        Returns list of files that should be analyzed for a LANG

        Given this object's TARGET, self.INCLUDE, and self.EXCLUDE will return list
        of all descendant files of directories in TARGET that end in extension
        typical for LANG. If self.INCLUDES is non empty then all files will have an ancestor
        that matches a pattern in self.INCLUDES. Will not include any file that has
        an ancestor that matches a pattern in self.EXCLUDES. Any explicitly named files
        in TARGET will bypass this global INCLUDE/EXCLUDE filter. The local INCLUDE/EXCLUDE
        filter is then applied.
        """
        paths = self.get_files_for_language(lang)

        paths = self.filter_includes(rule_includes, candidates=paths.kept)
        self.ignore_log.rule_includes[rule_id].update(paths.removed)

        paths = self.filter_excludes(rule_excludes, candidates=paths.kept)
        self.ignore_log.rule_excludes[rule_id].update(paths.removed)

        return paths.kept



import base64
import collections
import json
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional

from defusedxml import ElementTree as ET  # type: ignore
from packaging.version import InvalidVersion
from packaging.version import Version

from semgrep.error import SemgrepError
from semgrep.verbose_logging import getLogger

# NOTE: Defused XML doesn't export types :(


logger = getLogger(__name__)

from dependencyparser.models import LockfileDependency
from dependencyparser.models import PackageManagers


def extract_npm_lockfile_hash(s: str) -> Dict[str, List[str]]:
    """
    Go from:
        sha512-aePbxDmcYW++PaqBsJ+HYUFwCdv4LVvdnhBy78E57PIor8/OVvhMrADFFEDh8DHDFRv/O9i3lPhsENjO7QX0+A==
    To:
        sha512,
    """
    algorithm = s.split("-")[0]
    rest = s[len(algorithm) + 1 :]
    decode_base_64 = base64.b64decode(rest)
    return {algorithm: [base64.b16encode(decode_base_64).decode("ascii").lower()]}


def parse_Yarnlock_str(lockfile_text: str) -> Generator[LockfileDependency, None, None]:
    def extract_yarn_name(line: str) -> str:
        """
        "@babel/code-frame@^7.0.0", "@babel/code-frame@^7.10.4", "@babel/code-frame@^7.8.3":
        should parse as @babel/code-frame

        "@babel/types@^7.0.0", "@babel/types@^7.10.4", "@babel/types@^7.10.5", "@babel/types@^7.11.0", "@babel/types@^7.3.0", "@babel/types@^7.3.3", "@babel/types@^7.4.0", "@babel/types@^7.4.4", "@babel/types@^7.7.0", "@babel/types@^7.9.0":
        should parse as @babel/types

        lodash@4.17.18:
        should parse as lodash
        """
        if '"' not in line:
            return line.split("@")[0]
        else:
            first_quoted = "".join(line.split('"')[1:2])
            parsed_name = "@".join(first_quoted.split("@")[:-1])
            return parsed_name

    def remove_trailing_octothorpe(s: str) -> str:
        return "#".join(s.split("#")[:-1]) if "#" in s else s

    def remove_quotes(s: str) -> str:
        return s[1:-1]

    _comment, all_deps_text = lockfile_text.split("\n\n\n")
    dep_texts = all_deps_text.split("\n\n")
    for dep_text in dep_texts:
        lines = dep_text.split("\n")
        package_name = extract_yarn_name(lines[0])
        version = remove_quotes(lines[1].split()[1])
        resolved = remove_trailing_octothorpe(
            remove_quotes(lines[2].split()[1]).strip()
        )
        integrity = (
            lines[3].split()[1]
            if len(lines) > 3 and lines[3].strip().startswith("integrity")
            else None
        )
        yield LockfileDependency(
            package_name,
            version,
            PackageManagers.NPM,
            allowed_hashes=extract_npm_lockfile_hash(integrity) if integrity else {},
            resolved_url=[resolved],
        )


def parse_NPM_package_lock_str(
    lockfile_text: str,
) -> Generator[LockfileDependency, None, None]:
    as_json = json.loads(lockfile_text)
    # Newer versions of NPM (>= v7) use 'packages'
    # But 'dependencies' is kept up to date, and 'packages' uses relative, not absolute names
    # https://docs.npmjs.com/cli/v8/configuring-npm/package-lock-json
    if "dependencies" in as_json:
        deps = as_json["dependencies"]
    elif "packages" in as_json:
        deps = as_json["packages"]
    else:
        logger.info("Found package-lock with no 'dependencies' or 'packages'")
        return

    for dep, dep_blob in deps.items():
        version = dep_blob.get("version")
        if not version:
            logger.info(f"no version for dependency: {dep}")
            continue
        try:
            Version(version)
        # Version was a github commit
        except InvalidVersion:
            logger.info(f"no version for dependency: {dep}")
            continue

        resolved_url = dep_blob.get("resolved")
        integrity = dep_blob.get("integrity")
        yield LockfileDependency(
            dep,
            version,
            PackageManagers.NPM,
            allowed_hashes=extract_npm_lockfile_hash(integrity) if integrity else {},
            resolved_url=[resolved_url] if resolved_url else None,
        )


def parse_Pipfile_str(lockfile_text: str) -> Generator[LockfileDependency, None, None]:
    def extract_pipfile_hashes(
        hashes: List[str],
    ) -> Dict[str, List[str]]:
        output = collections.defaultdict(list)
        for h in hashes:
            algorithm = h.split(":")[0]
            rest = h[len(algorithm) + 1 :]  # pipfile is already in base16
            output[algorithm].append(rest.lower())
        return output

    def parse_dependency_blob(
        root_blob: Dict[str, Any]
    ) -> Generator[LockfileDependency, None, None]:
        for dep in root_blob:
            dep_blob = root_blob[dep]
            version = dep_blob.get("version", None)
            if not version:
                logger.info(f"no version for dependency: {dep}")
            else:
                version = version.replace("==", "")
                yield LockfileDependency(
                    dep,
                    version,
                    PackageManagers.PYPI,
                    resolved_url=None,
                    allowed_hashes=extract_pipfile_hashes(dep_blob["hashes"]),
                )

    as_json = json.loads(lockfile_text)
    yield from parse_dependency_blob(as_json["default"])
    develop_deps = as_json.get("develop")
    if develop_deps is not None:
        yield from parse_dependency_blob(develop_deps)


def parse_Gemfile_str(lockfile_text: str) -> Generator[LockfileDependency, None, None]:
    def parse_dep(s: str) -> LockfileDependency:
        # s == "    $DEP ($VERSION)"
        dep, paren_version = s.strip().split(" ")
        version = paren_version[1:-1]
        return LockfileDependency(
            dep, version, PackageManagers.GEM, resolved_url=None, allowed_hashes={}
        )

    lines = lockfile_text.split("\n")
    # No dependencies specified
    if "GEM" not in lines:
        yield from []
    GEM_idx = lines.index("GEM") + 1
    GEM_end_idx = lines[GEM_idx:].index(
        ""
    )  # A line with a single \n becomes the empty string upon splitting by \n
    all_deps = lines[GEM_idx:GEM_end_idx]
    yield from (
        parse_dep(dep) for dep in all_deps if dep[:4] == " " * 4 and dep[5] != " "
    )


def parse_Go_sum_str(lockfile_text: str) -> Generator[LockfileDependency, None, None]:
    # We currently ignore the +incompatible flag, pseudo versions, and the difference between a go.mod and a direct download
    def parse_dep(s: str) -> LockfileDependency:
        dep, version, hash = s.split()
        # drop 'v'
        version = version[1:]

        # drop /go.mod
        if "/" in version:
            version = version[: version.index("/")]
        # drop +incompatible
        if "+" in version:
            version = version[: version.index("+")]

        # drop pseudo version
        if "-" in version:
            version = version[: version.index("-")]

        # drop h1: and =
        hash = hash[3:-1]
        return LockfileDependency(
            dep,
            version,
            PackageManagers.GOMOD,
            # go.sum dep names are already URLs
            resolved_url=[dep],
            allowed_hashes={"gomod": [hash]},
        )

    lines = lockfile_text.split("\n")
    yield from (parse_dep(dep) for dep in lines)


def parse_Cargo_str(lockfile_text: str) -> Generator[LockfileDependency, None, None]:
    def parse_dep(s: str) -> LockfileDependency:
        lines = s.split("\n")[1:]
        dep = lines[0].split("=")[1].strip()[1:-1]
        version = lines[1].split("=")[1].strip()[1:-1]
        hash = lines[3].split("=")[1].strip()[1:-1]
        return LockfileDependency(
            dep,
            version,
            PackageManagers.CARGO,
            resolved_url=None,
            allowed_hashes={"sha256": [hash]},
        )

    deps = lockfile_text.split("[[package]]")[1:]
    yield from (parse_dep(dep) for dep in deps)


def parse_Pom_str(manifest_text: str) -> Generator[LockfileDependency, None, None]:
    NAMESPACE = "{http://maven.apache.org/POM/4.0.0}"

    def parse_dep(
        properties: Any,
        # Optional[ET.Element]
        el: Any
        # ET.Element
    ) -> Optional[LockfileDependency]:
        dep_el = el.find(f"{NAMESPACE}artifactId")
        if dep_el is None:
            return None
        dep = dep_el.text
        if dep is None:
            return None
        version_el = el.find(f"{NAMESPACE}version")
        if version_el is None:
            return None
        version = version_el.text
        if version is None:
            return None
        if version[0] == "$":
            if properties is None:
                raise SemgrepError("invalid pom.xml?")

            version = version[2:-1]
            prop_version = properties.find(f"{NAMESPACE}{version}")
            if prop_version is None:
                return None
            version = prop_version.text
            if version is None:
                return None

        try:
            # pom.xml does not specify an exact version, so we give up
            Version(version)
        except InvalidVersion:
            return None

        return LockfileDependency(
            dep, version, PackageManagers.MAVEN, resolved_url=None, allowed_hashes={}
        )

    root = ET.fromstring(manifest_text)
    deps = root.find(f"{NAMESPACE}dependencies")
    if deps is None:
        raise SemgrepError("No dependencies in pom.xml?")
    properties = root.find(f"{NAMESPACE}properties")
    for dep in deps:
        dep_opt = parse_dep(properties, dep)
        if dep_opt:
            yield dep_opt


LOCKFILE_PARSERS = {
    "pipfile.lock": parse_Pipfile_str,  # Python
    "yarn.lock": parse_Yarnlock_str,  # JavaScript
    "package-lock.json": parse_NPM_package_lock_str,  # JavaScript
    "gemfile.lock": parse_Gemfile_str,  # Ruby
    "go.sum": parse_Go_sum_str,  # Go
    "cargo.lock": parse_Cargo_str,  # Rust
    "pom.xml": parse_Pom_str,  # Java
}


def parse_lockfile_str(
    lockfile_text: str, filepath_for_reference: Path
) -> Generator[LockfileDependency, None, None]:
    # coupling with the github action, which decides to send files with these names back to us
    filepath = filepath_for_reference.name.lower()
    if filepath in LOCKFILE_PARSERS:
        return LOCKFILE_PARSERS[filepath](lockfile_text)
    else:
        raise SemgrepError(
            f"don't know how to parse this filename: {filepath_for_reference}"
        )



"""
This file encapsulates classes necessary in parsing semgrep-core json output into a typed object

The objects in this class should expose functionality that returns objects that the rest of the codebase
interacts with (e.g. the rest of the codebase should be interacting with RuleMatch objects instead of CoreMatch
and SemgrepCoreError instead of CoreError objects).

The precise type of the response from semgrep-core is specified in
Semgrep_core_response.atd, currently at:
https://github.com/returntocorp/semgrep/blob/develop/semgrep-core/src/core-response/Semgrep_core_response.atd
"""
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import NewType
from typing import Optional
from typing import Set
from typing import Tuple

from attrs import define
from attrs import frozen

import semgrep.output_from_core as core
from semgrep.error import LegacySpan
from semgrep.error import Level
from semgrep.error import SemgrepCoreError
from semgrep.output_from_core import MetavarValue
from semgrep.output_from_core import SkipReason
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch
from semgrep.rule_match import RuleMatchSet
from semgrep.types import JsonObject
from semgrep.types import RuleId
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)

CoreErrorMessage = NewType("CoreErrorMessage", str)
CoreErrorType = NewType("CoreErrorType", str)
SkipDetails = NewType("SkipDetails", str)

CoreRulesParseTime = NewType("CoreRulesParseTime", float)


@frozen
class CoreMatch:
    """
    Encapsulates finding returned by semgrep-core
    """

    rule: Rule
    path: Path
    start: core.Position
    end: core.Position
    extra: Dict[str, Any]
    metavars: Dict[str, MetavarValue]

    @classmethod
    def make(cls, rule_table: Dict[str, Rule], match: core.Match) -> "CoreMatch":
        rule = rule_table[match.rule_id.value]
        path = Path(match.location.path)
        start = match.location.start
        end = match.location.end
        extra = match.extra.to_json()
        metavars = match.extra.metavars
        return cls(rule, path, start, end, extra, metavars)


@frozen
class CoreError:
    """
    Encapsulates error object returned by semgrep-core
    and handles conversion into SemgrepCoreError class that rest of codebase understands.
    """

    error_type: CoreErrorType
    rule_id: Optional[RuleId]
    path: Path
    start: core.Position
    end: core.Position
    message: CoreErrorMessage
    level: Level
    spans: Optional[Tuple[LegacySpan, ...]]
    details: Optional[str]

    @classmethod
    def make(cls, error: core.Error) -> "CoreError":
        error_type = CoreErrorType(error.error_type)
        rule_id = RuleId(error.rule_id.value) if error.rule_id else None
        path = Path(error.location.path)
        start = error.location.start
        end = error.location.end
        message = CoreErrorMessage(error.message)

        # Hackily convert the level string to Semgrep expectations
        level_str = error.severity.kind
        if level_str.upper() == "WARNING":
            level_str = "WARN"
        if level_str.upper() == "ERROR_":
            level_str = "ERROR"
        level = Level[level_str.upper()]
        details = error.details

        spans = None
        if error.yaml_path:
            yaml_path = tuple(error.yaml_path[::-1])
            spans = tuple([LegacySpan(start, end, yaml_path)])  # type: ignore
        return cls(
            error_type, rule_id, path, start, end, message, level, spans, details
        )

    def is_timeout(self) -> bool:
        """
        Return if this error is a match timeout
        """
        return self.error_type == CoreErrorType("Timeout")

    def to_semgrep_error(self) -> SemgrepCoreError:
        reported_rule_id = self.rule_id

        # TODO benchmarking code relies on error code value right now
        # See https://semgrep.dev/docs/cli-usage/ for meaning of codes
        if self.error_type == CoreErrorType(
            "Syntax error"
        ) or self.error_type == CoreErrorType("Lexical error"):
            code = 3
            reported_rule_id = None  # Rule id not important for parse errors
        else:
            code = 2

        return SemgrepCoreError(
            code,
            self.level,
            self.error_type,
            reported_rule_id,
            self.path,
            self.start,
            self.end,
            self.message,
            self.spans,
            self.details,
        )


@frozen
class CoreSkipped:
    rule_id: Optional[RuleId]
    path: Path
    reason: SkipReason
    details: SkipDetails

    @classmethod
    def make(cls, skipped: core.SkippedTarget) -> "CoreSkipped":
        rule_id = RuleId(skipped.rule_id.value) if skipped.rule_id else None
        path = Path(skipped.path)
        reason = skipped.reason
        details = SkipDetails(skipped.details)

        if rule_id:
            rule_info = f"rule {rule_id}"
        else:
            rule_info = "all rules"
        logger.verbose(f"skipped '{str(path)}' [{rule_info}]: {reason}: {details}")
        return cls(rule_id, path, reason, details)


@frozen
class CoreRuleTiming:  # For a given target
    rule: Rule
    parse_time: float
    match_time: float

    @classmethod
    def make(
        cls, rule_table: Dict[str, Rule], rule_time: core.RuleTimes
    ) -> "CoreRuleTiming":
        rule = rule_table[rule_time.rule_id]
        parse_time = rule_time.parse_time
        match_time = rule_time.match_time
        return cls(rule, parse_time, match_time)


@frozen
class CoreTargetTiming:
    target: Path
    per_rule_timings: List[CoreRuleTiming]
    run_time: float

    @classmethod
    def make(
        cls, rule_table: Dict[str, Rule], target_time: core.TargetTime
    ) -> "CoreTargetTiming":
        target = Path(target_time.path)
        per_rule_timings = [
            CoreRuleTiming.make(rule_table, rule_time)
            for rule_time in target_time.rule_times
        ]
        run_time = target_time.run_time
        return cls(target, per_rule_timings, run_time)


@frozen
class CoreTiming:
    rules: List[Rule]
    target_timings: List[CoreTargetTiming]
    rules_parse_time: CoreRulesParseTime

    @classmethod
    def make(
        cls, rule_table: Dict[str, Rule], time: Optional[core.Time]
    ) -> "CoreTiming":
        if not time:
            return cls([], [], CoreRulesParseTime(0.0))

        rules = [rule_table[rule] for rule in time.rules]
        target_timings = [
            CoreTargetTiming.make(rule_table, target) for target in time.targets
        ]
        rules_parse_time = CoreRulesParseTime(
            time.rules_parse_time if time.rules_parse_time else 0.0
        )
        return cls(rules, target_timings, rules_parse_time)


@define
class CoreOutput:
    """
    Parses output of semgrep-core
    """

    matches: List[CoreMatch]
    errors: List[CoreError]
    skipped: List[CoreSkipped]
    timing: CoreTiming

    @classmethod
    def parse(cls, rules: List[Rule], raw_json: JsonObject) -> "CoreOutput":
        rule_table = {rule.id: rule for rule in rules}

        match_results = core.MatchResults.from_json(raw_json)

        parsed_errors = [CoreError.make(error) for error in match_results.errors]
        parsed_matches = [
            CoreMatch.make(rule_table, match) for match in match_results.matches
        ]
        parsed_skipped = [
            CoreSkipped.make(skip) for skip in match_results.skipped_targets
        ]
        parsed_timings = CoreTiming.make(rule_table, match_results.time)

        return cls(parsed_matches, parsed_errors, parsed_skipped, parsed_timings)

    def rule_matches(self, rules: List[Rule]) -> Dict[Rule, List[RuleMatch]]:
        """
        Convert core_match objects into RuleMatch objects that the rest of the codebase
        interacts with.

        For now assumes that all matches encapsulated by this object are from the same rulee
        """

        def interpolate(text: str, metavariables: Dict[str, str]) -> str:
            """Interpolates a string with the metavariables contained in it, returning a new string"""

            # Sort by metavariable length to avoid name collisions (eg. $X2 must be handled before $X)
            for metavariable in sorted(metavariables.keys(), key=len, reverse=True):
                text = text.replace(metavariable, metavariables[metavariable])

            return text

        def read_metavariables(match: CoreMatch) -> Dict[str, str]:
            result = {}

            # open path and ignore non-utf8 bytes. https://stackoverflow.com/a/56441652
            with open(match.path, errors="replace") as fd:
                for metavariable, metavariable_data in match.metavars.items():
                    # Offsets are start inclusive and end exclusive
                    start_offset = metavariable_data.start.offset
                    end_offset = metavariable_data.end.offset
                    length = end_offset - start_offset

                    fd.seek(start_offset)
                    result[metavariable] = fd.read(length)

            return result

        def convert_to_rule_match(match: CoreMatch) -> RuleMatch:
            rule = match.rule
            metavariables = read_metavariables(match)
            message = interpolate(rule.message, metavariables)
            fix = interpolate(rule.fix, metavariables) if rule.fix else None

            return RuleMatch(
                rule._id,
                message=message,
                metadata=rule.metadata,
                severity=rule.severity,
                fix=fix,
                fix_regex=rule.fix_regex,
                path=match.path,
                start=match.start,
                end=match.end,
                extra=match.extra,
            )

        findings: Dict[Rule, RuleMatchSet] = {rule: RuleMatchSet() for rule in rules}
        seen_cli_unique_keys: Set[Tuple] = set()
        for match in self.matches:
            rule_match = convert_to_rule_match(match)
            if rule_match.cli_unique_key in seen_cli_unique_keys:
                continue
            seen_cli_unique_keys.add(rule_match.cli_unique_key)
            findings[match.rule].add(rule_match)

        # Sort results so as to guarantee the same results across different
        # runs. Results may arrive in a different order due to parallelism
        # (-j option).
        return {rule: sorted(matches) for rule, matches in findings.items()}



import os
import sys
import time
import uuid
from typing import NoReturn
from typing import Optional
from typing import Tuple

import click

from semgrep.app import app_session
from semgrep.app import auth
from semgrep.commands.wrapper import handle_command_errors
from semgrep.constants import IN_DOCKER
from semgrep.constants import IN_GH_ACTION
from semgrep.constants import SEMGREP_URL
from semgrep.error import FATAL_EXIT_CODE
from semgrep.settings import SETTINGS
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)


def make_login_url() -> Tuple[uuid.UUID, str]:
    session_id = uuid.uuid4()
    return (
        session_id,
        f"{SEMGREP_URL}/login?cli-token={session_id}&docker={IN_DOCKER}&gha={IN_GH_ACTION}",
    )


@click.command()
@handle_command_errors
def login() -> NoReturn:
    """
    Obtain and save credentials for semgrep.dev

    Looks for an semgrep.dev API token in the environment variable SEMGREP_API_TOKEN_SETTINGS_KEY.
    If not defined and running in a TTY, prompts interactively.
    Once token is found, saves it to global settings file
    """
    saved_login_token = auth._read_token_from_settings_file()
    if saved_login_token:
        click.echo(
            f"API token already exists in {SETTINGS.get_path_to_settings()}. To login with a different token logout use `semgrep logout`"
        )
        sys.exit(FATAL_EXIT_CODE)

    # If the token is provided as an environment variable, save it to the settings file.
    env_var_token = os.environ.get(auth.SEMGREP_LOGIN_TOKEN_ENVVAR_NAME)
    if env_var_token is not None and len(env_var_token) > 0:
        if not save_token(env_var_token, echo_token=False):
            sys.exit(FATAL_EXIT_CODE)
        sys.exit(0)

    # If token doesn't already exist in the settings file or as an environment variable,
    # interactively prompt the user to supply it (if we are in a TTY).
    if not auth.is_a_tty():
        click.echo(
            f"Error: semgrep login is an interactive command: run in an interactive terminal (or define {auth.SEMGREP_LOGIN_TOKEN_ENVVAR_NAME})",
            err=True,
        )
        sys.exit(FATAL_EXIT_CODE)

    session_id, url = make_login_url()
    click.echo(
        "Login enables additional proprietary Semgrep Registry rules and running custom policies from Semgrep App."
    )
    click.echo(f"Login at: {url}")
    click.echo(
        "\nOnce you've logged in, return here and you'll be ready to start using new Semgrep rules."
    )
    WAIT_BETWEEN_RETRY_IN_SEC = 6  # So every 10 retries is a minute
    MAX_RETRIES = 30  # Give users 3 minutes to log in / open link

    for _ in range(MAX_RETRIES):
        r = app_session.post(
            f"{SEMGREP_URL}/api/agent/tokens/requests",
            json={"token_request_key": str(session_id)},
        )
        if r.status_code == 200:
            as_json = r.json()
            if save_token(as_json.get("token"), echo_token=True):
                sys.exit(0)
            else:
                sys.exit(FATAL_EXIT_CODE)
        elif r.status_code != 404:
            click.echo(
                f"Unexpected failure from {SEMGREP_URL}: status code {r.status_code}; please contact support@r2c.dev if this persists",
                err=True,
            )

        time.sleep(WAIT_BETWEEN_RETRY_IN_SEC)

    click.echo(
        f"Failed to login: please check your internet connection or contact support@r2c.dev",
        err=True,
    )
    sys.exit(FATAL_EXIT_CODE)


def save_token(login_token: Optional[str], echo_token: bool) -> bool:
    if login_token is not None and auth.is_valid_token(login_token):
        auth.set_token(login_token)
        click.echo(
            f"Saved login token\n\n\t{login_token if echo_token else '<redacted>'}\n\nin {SETTINGS.get_path_to_settings()}."
        )
        click.echo(
            f"Note: You can always generate more tokens at {SEMGREP_URL}/orgs/-/settings/tokens"
        )
        return True
    else:
        click.echo("Login token is not valid. Please try again.", err=True)
        return False


@click.command()
@handle_command_errors
def logout() -> None:
    """
    Remove locally stored credentials to semgrep.dev
    """
    auth.delete_token()
    click.echo("Logged out (log back in with `semgrep login`)")



import json
from pathlib import Path
from typing import Collection
from typing import Mapping
from typing import NewType
from typing import Optional

from attrs import frozen

from semgrep.error import UnknownLanguageError
from semgrep.rule_lang import Span
from semgrep.types import JsonObject

Mode = NewType("Mode", str)
FileExtension = NewType("FileExtension", str)
Shebang = str

JOIN_MODE = Mode("join")
SEARCH_MODE = DEFAULT_MODE = Mode("search")


class Language(str):
    @property
    def definition(self) -> "LanguageDefinition":
        return LANGUAGE.definition_by_id[self]


@frozen
class LanguageDefinition:
    """
    Mirrors schema of lang.json (see lang/README.md) for each language
    """

    id: Language
    name: str
    keys: Collection[str]
    exts: Collection[FileExtension]
    reverse_exts: Collection[str]
    shebangs: Collection[Shebang]

    @classmethod
    def from_dict(cls, data: JsonObject) -> "LanguageDefinition":
        return cls(
            id=Language(data["id"]),
            name=data["name"],
            keys=data["keys"],
            exts=data["exts"],
            reverse_exts=data.get("reverse_exts", data["exts"]),
            shebangs=data.get("shebangs", []),
        )


class _LanguageData:
    def __init__(self) -> None:
        with (Path(__file__).parent / "lang" / "lang.json").open() as fd:
            data = json.load(fd)

        self.definition_by_id: Mapping[Language, LanguageDefinition] = {
            Language(d["id"]): LanguageDefinition.from_dict(d) for d in data
        }
        self.lang_by_key: Mapping[str, Language] = {
            key.lower(): lang
            for lang, definition in self.definition_by_id.items()
            for key in definition.keys
        }
        self.lang_by_ext: Mapping[str, Language] = {
            ext: lang
            for lang, definition in self.definition_by_id.items()
            for ext in definition.reverse_exts
        }
        self.all_language_keys: Collection[str] = sorted(self.lang_by_key.keys())

    def resolve(self, lang_str: str, span: Optional[Span] = None) -> Language:
        """
        Convert an inputted string representing a language to a Language

        :param lang_str: string representing a language (e.g. "C#")
        :param span: span of language string in the original file (for error reporting),
                    None if resolve was called within semgrep
        """
        normalized = lang_str.lower()
        if normalized in self.lang_by_key:
            return self.lang_by_key[normalized]
        else:
            spans = [span.with_context(before=1, after=1)] if span else []
            raise UnknownLanguageError(
                short_msg=f"invalid language: {normalized}",
                long_msg=f"unsupported language: {normalized}. {self.show_suppported_languages_message()}",
                spans=spans,
            )

    def show_suppported_languages_message(self) -> str:
        return f"supported languages are: {', '.join(self.all_language_keys)}"


LANGUAGE = _LanguageData()


ALLOWED_GLOB_TYPES = ("include", "exclude")



import os
from typing import Any
from typing import Optional
from typing import Set

import requests
import urllib3
from attrs import define
from attrs import field

from semgrep import __VERSION__
from semgrep.app import auth


@define
class UserAgent:
    """
    Generates the user agent string we send to Semgrep App.

    >>> from semgrep.app import app_session
    >>> str(app_session.user_agent)
    'semgrep/0.1.2'
    >>> app_session.user_agent.tags.add("testing")
    >>> str(app_session.user_agent)
    'semgrep/0.1.2 (testing)'
    """

    name: str = field(default=f"Semgrep", init=False)
    version: str = field(default=__VERSION__, init=False)
    tags: Set[str] = field(init=False)

    @tags.default
    def get_default_tags(self) -> Set[str]:
        if os.getenv("SEMGREP_USER_AGENT_APPEND"):
            return set({os.environ["SEMGREP_USER_AGENT_APPEND"]})
        return set()

    def __str__(self) -> str:
        result = f"{self.name}/{self.version}"
        for note in sorted(self.tags):
            result += f" ({note})"
        return result


class AppSession(requests.Session):
    """
    Send requests to Semgrep App with this session.

    Use the instance at `semgrep.app.app_session` instead of creating a new one.

    The following features are added over the base Session class:
    - A default retrying policy is added to each request
    - A User-Agent is automatically added to each request
    - A default timeout of 30 seconds is added to each request
    - If a token is available, it is added to the request as an Authorization header

    Normal usage:
    >>> from semgrep.app import app_session
    >>> app_session.get(url)

    Disable custom user agent for a request:
    >>> app_session.get(url, headers={"User-Agent": None}))

    Disable timeout for a request:
    >>> app_session.get(url, timeout=None)

    Disable authentication for a request:
    >>> app_session.get(url, headers={"Authorization": None})
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.user_agent = UserAgent()
        self.token: Optional[str] = None

        # retry after 4, 8, 16 seconds
        retry_adapter = requests.adapters.HTTPAdapter(
            max_retries=urllib3.Retry(
                total=3,
                backoff_factor=4,
                allowed_methods=["GET", "POST"],
                status_forcelist=(413, 429, 500, 502, 503),
            ),
        )
        self.mount("https://", retry_adapter)
        self.mount("http://", retry_adapter)

    def authenticate(self) -> None:
        self.token = auth.get_token()

    def request(self, *args: Any, **kwargs: Any) -> requests.Response:
        kwargs.setdefault("timeout", 30)
        kwargs.setdefault("headers", {})
        kwargs["headers"].setdefault("User-Agent", str(self.user_agent))
        if self.token:
            kwargs["headers"].setdefault("Authorization", f"Bearer {self.token}")
        return super().request(*args, **kwargs)



import logging
import os
import sys
from typing import Optional

from semgrep.constants import SEMGREP_URL
from semgrep.settings import SETTINGS

logger = logging.getLogger(__name__)

SEMGREP_LOGIN_TOKEN_ENVVAR_NAME = "SEMGREP_APP_TOKEN"
SEMGREP_API_TOKEN_SETTINGS_KEY = "api_token"


def is_valid_token(token: str) -> bool:
    """
    Returns true if token is valid
    """
    from semgrep.app import app_session  # avoiding circular imports

    r = app_session.get(
        f"{SEMGREP_URL}/api/agent/deployments/current",
        headers={"Authorization": f"Bearer {token}"},
    )
    return r.ok


def get_deployment_id() -> Optional[int]:
    """
    Returns the deployment_id attached to an api_token as int

    Returns None if api_token is invalid/doesn't have associated deployment
    """
    from semgrep.app import app_session  # avoiding circular imports

    r = app_session.get(f"{SEMGREP_URL}/api/agent/deployments/current")

    if r.ok:
        data = r.json()
        return data.get("deployment", {}).get("id")  # type: ignore
    else:
        return None


def get_token() -> Optional[str]:
    """
    Get saved token in following order:
    - env var
    - settings file
    - None
    """
    login_token = os.environ.get(SEMGREP_LOGIN_TOKEN_ENVVAR_NAME)
    if login_token is not None:
        logger.debug(
            f"Using environment variable {SEMGREP_LOGIN_TOKEN_ENVVAR_NAME} as api token"
        )
        return login_token

    return _read_token_from_settings_file()


def _read_token_from_settings_file() -> Optional[str]:
    """
    Read api token from settings file

    Returns None if api token not in settings file
    """
    logger.debug("Getting API token from settings file")
    login_token = SETTINGS.get_setting(SEMGREP_API_TOKEN_SETTINGS_KEY, default=None)

    if login_token and not isinstance(login_token, str):
        raise ValueError()

    return login_token


def set_token(token: str) -> None:
    """
    Save api token to settings file
    """
    logger.debug("Saving API token in settings file")
    SETTINGS.add_setting(SEMGREP_API_TOKEN_SETTINGS_KEY, token)


def delete_token() -> None:
    """
    Remove api token from settings file
    """
    logger.debug("Deleting api token from settings file")
    SETTINGS.delete_setting(SEMGREP_API_TOKEN_SETTINGS_KEY)


def is_a_tty() -> bool:
    """
    Whether or not the terminal is interactive (a tty)
    Separated out to make test mocking easier
    """
    return sys.stderr.isatty()



import hashlib
import json
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Mapping
from typing import Sequence

from semgrep.constants import RuleSeverity
from semgrep.error import SemgrepError
from semgrep.formatter.base import BaseFormatter
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatch


def _to_gitlab_severity(semgrep_severity: RuleSeverity) -> str:
    # Todo: Semgrep states currently don't map super well to Gitlab schema.
    conversion_table: Dict[RuleSeverity, str] = {
        RuleSeverity.INFO: "Info",
        RuleSeverity.WARNING: "Medium",
        RuleSeverity.ERROR: "High",
    }
    return conversion_table.get(semgrep_severity, "Unknown")


def _construct_semgrep_rule_url(rule_id: str) -> str:
    # this is a hack to fix name -> registry disagreement
    components = rule_id.split(".")
    result = []
    for chunk in components:
        if chunk not in result:
            result.append(chunk)
    rule_name = ".".join(result)
    return f"https://semgrep.dev/r/{rule_name}"


class GitlabSastFormatter(BaseFormatter):
    def _format_rule_match(self, rule_match: RuleMatch) -> Mapping[str, Any]:
        # create UUID from sha256 hash
        return {
            "id": str(rule_match.uuid),
            "category": "sast",
            # CVE is a required field from Gitlab schema.
            # It also is part of the determination for uniqueness
            # of a detected secret
            # /regardless/ of differentiated ID. See issue 262648.
            # https://gitlab.com/gitlab-org/gitlab/-/issues/262648
            # Gitlab themselves mock a CVE for findings that lack a CVE
            # Format: path:hash-of-file-path:check_id
            "cve": "{}:{}:{}".format(
                rule_match.path,
                hashlib.sha256(bytes(rule_match.path)).hexdigest(),
                rule_match.rule_id,
            ),
            "message": rule_match.message,
            "severity": _to_gitlab_severity(rule_match.severity),
            # Semgrep is designed to be a low-FP tool by design.
            # Does hard-coding confidence make sense here?
            "confidence": "High",
            "scanner": {
                "id": "semgrep",
                "name": "Semgrep",
                "vendor": {"name": "Semgrep"},
            },
            "location": {
                "file": str(rule_match.path),
                # Gitlab only uses line identifiers
                "start_line": rule_match.start.line,
                "end_line": rule_match.end.line,
                "dependency": {"package": {}},
            },
            "identifiers": [
                {
                    "type": "semgrep_type",
                    "name": f"Semgrep - {rule_match.rule_id}",
                    "value": rule_match.rule_id,
                    "url": _construct_semgrep_rule_url(rule_match.rule_id),
                }
            ],
        }

    def format(
        self,
        rules: Iterable[Rule],
        rule_matches: Iterable[RuleMatch],
        semgrep_structured_errors: Sequence[SemgrepError],
        extra: Mapping[str, Any],
    ) -> str:
        """
        Format matches in GitLab SAST report compliant JSON.

        - Written based on:
            https://github.com/returntocorp/semgrep-action/blob/678eff1a4269ed04b76631771688c8be860ec4e9/src/semgrep_agent/findings.py#L137-L165
        - Docs:
            https://docs.gitlab.com/ee/user/application_security/sast/#reports-json-format
        - Schema:
            https://gitlab.com/gitlab-org/security-products/security-report-schemas/-/blob/master/dist/sast-report-format.json
        """
        output_dict = {
            "$schema": "https://gitlab.com/gitlab-org/security-products/security-report-schemas/-/blob/master/dist/sast-report-format.json",
            "version": "2.0",
            "vulnerabilities": [
                self._format_rule_match(rule_match)
                for rule_match in rule_matches
                if rule_match.severity != RuleSeverity.INVENTORY
            ],
        }

        return json.dumps(output_dict, sort_keys=True, indent=2)



import os
import subprocess
import sys
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import click

import semgrep.semgrep_main
from semgrep.app import app_session
from semgrep.app import auth
from semgrep.app.metrics import metric_manager
from semgrep.app.scans import ScanHandler
from semgrep.commands.scan import CONTEXT_SETTINGS
from semgrep.commands.scan import scan_options
from semgrep.commands.wrapper import handle_command_errors
from semgrep.constants import OutputFormat
from semgrep.constants import SEMGREP_URL
from semgrep.error import FATAL_EXIT_CODE
from semgrep.error import INVALID_API_KEY_EXIT_CODE
from semgrep.error import SemgrepError
from semgrep.git import GIT_SH_TIMEOUT
from semgrep.ignores import IGNORE_FILE_NAME
from semgrep.meta import generate_meta_from_environment
from semgrep.meta import GithubMeta
from semgrep.meta import GitMeta
from semgrep.output import OutputHandler
from semgrep.output import OutputSettings
from semgrep.rule import Rule
from semgrep.rule_match import RuleMatchMap
from semgrep.types import MetricsState
from semgrep.util import set_flags
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)

# These patterns are excluded via --exclude regardless of other ignore configuration
ALWAYS_EXCLUDE_PATTERNS = [".semgrep/", ".semgrep_logs/"]

# These patterns are excluded via --exclude unless the user provides their own .semgrepignore
DEFAULT_EXCLUDE_PATTERNS = ["test/", "tests/", "*_test.go"]


def yield_valid_patterns(patterns: Iterable[str]) -> Iterable[str]:
    """
    Parses patterns from semgrep.dev and returns the lines that
    are non-empty and do not start with #
    """
    for pattern in patterns:
        pattern = pattern.strip()

        if not pattern:
            continue
        if pattern.startswith("#"):
            continue

        yield pattern


def yield_exclude_paths(requested_patterns: Sequence[str]) -> Iterable[str]:
    patterns = [*yield_valid_patterns(requested_patterns), *ALWAYS_EXCLUDE_PATTERNS]
    if Path(IGNORE_FILE_NAME).is_file() and not requested_patterns:
        patterns.extend(DEFAULT_EXCLUDE_PATTERNS)

    yield from patterns


@contextmanager
def fix_head_if_github_action(metadata: GitMeta) -> Iterator[None]:
    """
    GHA can checkout the incorrect commit for a PR (it will create a fake merge commit),
    so we need to reset the head to the actual PR branch head before continuing.

    Assumes cwd is a valid git project and that if we are in github-actions pull_request,
    that metadata.base_commit_ref and metadata.head_ref point are the head commits of
    the target merge branch and current branch respectively
    """
    if isinstance(metadata, GithubMeta) and metadata.is_pull_request_event:
        assert metadata.head_ref is not None  # Not none when github action PR

        logger.info("Fixing git state for github action pull request")
        logger.debug(
            f"base_commit_ref: {metadata.base_commit_ref}, head_ref: {metadata.head_ref}"
        )
        logger.debug("Calling git rev-parse HEAD")
        rev_parse = subprocess.run(
            ["git", "rev-parse", "HEAD"],
            encoding="utf-8",
            check=True,
            timeout=GIT_SH_TIMEOUT,
            capture_output=True,
        )
        logger.debug(f"git rev-parse stdout: {rev_parse.stdout}")
        logger.debug(f"git rev-parse stderr: {rev_parse.stderr}")
        stashed_rev = rev_parse.stdout.rstrip()
        logger.debug(f"stashed_rev: {stashed_rev}")

        logger.debug(f"Not on head ref {metadata.head_ref}; checking that out now.")
        checkout = subprocess.run(
            ["git", "checkout", metadata.head_ref],
            encoding="utf-8",
            check=True,
            capture_output=True,
            timeout=GIT_SH_TIMEOUT,
        )
        logger.debug(f"git checkout stdout: {checkout.stdout}")
        logger.debug(f"git checkout stderr: {checkout.stderr}")

        try:
            yield
        finally:
            logger.info(f"Returning to original head revision {stashed_rev}")
            subprocess.run(
                ["git", "checkout", stashed_rev],
                encoding="utf-8",
                capture_output=True,
                check=True,
                timeout=GIT_SH_TIMEOUT,
            )
    else:
        # Do nothing
        yield


@click.command(context_settings=CONTEXT_SETTINGS)
@click.pass_context
@scan_options
@click.option(
    "--audit-on",
    envvar="SEMGREP_AUDIT_ON",
    multiple=True,
    type=str,
    hidden=True,
)
@click.option(
    "--config",
    "-c",
    "-f",
    multiple=True,
    help="""
        YAML configuration file, directory of YAML files ending in
        .yml|.yaml, URL of a configuration file, or Semgrep registry entry name.
        \n\n
        Use --config auto to automatically obtain rules tailored to this project; your project URL will be used to log in
         to the Semgrep registry.
        \n\n
        To run multiple rule files simultaneously, use --config before every YAML, URL, or Semgrep registry entry name.
         For example `semgrep --config p/python --config myrules/myrule.yaml`
        \n\n
        See https://semgrep.dev/docs/writing-rules/rule-syntax for information on configuration file format.
    """,
    envvar="SEMGREP_RULES",
)
@click.option(
    "--dry-run",
    is_flag=True,
    help="""
        When set, will not start a scan on semgrep.dev and will not report findings.
        Instead will print out json objects it would have sent.
    """,
)
@handle_command_errors
def ci(
    ctx: click.Context,
    *,
    audit_on: Sequence[str],
    autofix: bool,
    baseline_commit: Optional[str],
    config: Optional[Tuple[str, ...]],
    debug: bool,
    dry_run: bool,
    emacs: bool,
    enable_nosem: bool,
    enable_version_check: bool,
    exclude: Optional[Tuple[str, ...]],
    force_color: bool,
    gitlab_sast: bool,
    gitlab_secrets: bool,
    include: Optional[Tuple[str, ...]],
    jobs: int,
    json: bool,
    junit_xml: bool,
    max_chars_per_line: int,
    max_lines_per_finding: int,
    max_memory: int,
    max_target_bytes: int,
    metrics: Optional[MetricsState],
    metrics_legacy: Optional[MetricsState],
    optimizations: str,
    output: Optional[str],
    sarif: bool,
    quiet: bool,
    rewrite_rule_ids: bool,
    scan_unknown_extensions: bool,
    time_flag: bool,
    timeout_threshold: int,
    timeout: int,
    use_git_ignore: bool,
    verbose: bool,
    vim: bool,
) -> None:
    """
    The recommended way to run semgrep in CI

    In pull_request/merge_request (PR/MR) contexts, `semgrep ci` will only report findings
    that were introduced by the PR/MR.

    When logged in, `semgrep ci` runs rules configured on Semgrep App and sends findings
    to your findings dashboard.

    Only displays findings that were marked as blocking.
    """

    set_flags(verbose=verbose, debug=debug, quiet=quiet, force_color=force_color)

    metric_manager.configure(metrics, metrics_legacy)
    scan_handler = None

    if not app_session.token and not config:
        # Not logged in and no explicit config
        logger.info("run `semgrep login` before using `semgrep ci` or set `--config`")
        sys.exit(INVALID_API_KEY_EXIT_CODE)
    elif not app_session.token and config:
        # Not logged in but has explicit config
        logger.info(f"Running `semgrep ci` without API token but with configs {config}")
    elif app_session.token and config:
        # Logged in but has explicit config
        logger.info(
            "Cannot run `semgrep ci` while logged in and with explicit config. Use semgrep.dev to configure rules to run."
        )
        sys.exit(FATAL_EXIT_CODE)
    else:
        assert app_session.token  # Must be defined here
        if not auth.is_valid_token(app_session.token):
            logger.info(
                "API token not valid. Try to run `semgrep logout` and `semgrep login` again.",
            )
            sys.exit(INVALID_API_KEY_EXIT_CODE)
        scan_handler = ScanHandler(dry_run)

    output_format = OutputFormat.TEXT
    if json:
        output_format = OutputFormat.JSON
    elif gitlab_sast:
        output_format = OutputFormat.GITLAB_SAST
    elif gitlab_secrets:
        output_format = OutputFormat.GITLAB_SECRETS
    elif junit_xml:
        output_format = OutputFormat.JUNIT_XML
    elif sarif:
        output_format = OutputFormat.SARIF
    elif emacs:
        output_format = OutputFormat.EMACS
    elif vim:
        output_format = OutputFormat.VIM

    output_settings = OutputSettings(
        output_format=output_format,
        output_destination=output,
        verbose_errors=verbose,
        timeout_threshold=timeout_threshold,
        output_time=time_flag,
        output_per_finding_max_lines_limit=max_lines_per_finding,
        output_per_line_max_chars_limit=max_chars_per_line,
    )
    output_handler = OutputHandler(output_settings)
    metadata = generate_meta_from_environment(baseline_commit)

    logger.info("Scan environment:")
    logger.info(
        f"  versions    - semgrep {semgrep.__VERSION__} on python {sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
    )
    logger.info(
        f"  environment - running in environment {metadata.environment}, triggering event is {metadata.event_name}"
    )
    to_server = "" if SEMGREP_URL == "https://semgrep.dev" else f" to {SEMGREP_URL}"
    if scan_handler:
        logger.info(
            f"  semgrep.dev - authenticated{to_server} as {scan_handler.deployment_name}"
        )
    logger.info("")

    try:
        with fix_head_if_github_action(metadata):
            try:
                logger.info("Fetching configuration from semgrep.dev")
                # Note this needs to happen within fix_head_if_github_action
                # so that metadata of current commit is correct
                if scan_handler:
                    scan_handler.start_scan(metadata.to_dict())
                    config = (scan_handler.scan_rules_url,)
            except Exception as e:
                import traceback

                traceback.print_exc()
                logger.info(f"Could not start scan {e}")
                sys.exit(FATAL_EXIT_CODE)

            # Append ignores configured on semgrep.dev
            requested_excludes = scan_handler.ignore_patterns if scan_handler else []
            if requested_excludes:
                logger.info(
                    f"Adding ignore patterns configured on semgrep.dev as `--exclude` options: {exclude}"
                )

            assert exclude is not None  # exclude is default empty tuple
            exclude = (*exclude, *yield_exclude_paths(requested_excludes))

            assert config  # Config has to be defined here. Helping mypy out
            start = time.time()
            (
                filtered_matches_by_rule,
                semgrep_errors,
                all_targets,
                ignore_log,
                filtered_rules,
                profiler,
                profiling_data,
                shown_severities,
            ) = semgrep.semgrep_main.main(
                output_handler=output_handler,
                target=[os.curdir],  # semgrep ci only scans cwd
                pattern=None,
                lang=None,
                configs=config,
                no_rewrite_rule_ids=(not rewrite_rule_ids),
                jobs=jobs,
                include=include,
                exclude=exclude,
                max_target_bytes=max_target_bytes,
                autofix=scan_handler.autofix if scan_handler else False,
                dryrun=True,
                disable_nosem=True,
                no_git_ignore=(not use_git_ignore),
                timeout=timeout,
                max_memory=max_memory,
                timeout_threshold=timeout_threshold,
                skip_unknown_extensions=(not scan_unknown_extensions),
                optimizations=optimizations,
                baseline_commit=metadata.base_commit_ref,
            )
    except SemgrepError as e:
        output_handler.handle_semgrep_errors([e])
        output_handler.output({}, all_targets=set(), filtered_rules=[])
        logger.info(f"Encountered error when running rules: {e}")
        if isinstance(e, SemgrepError):
            exit_code = e.code
        else:
            exit_code = FATAL_EXIT_CODE
        if scan_handler:
            scan_handler.report_failure(exit_code)
        sys.exit(exit_code)

    total_time = time.time() - start

    # Split up rules into respective categories:
    blocking_rules: List[Rule] = []
    nonblocking_rules: List[Rule] = []
    cai_rules: List[Rule] = []
    for rule in filtered_rules:
        if "r2c-internal-cai" in rule.id:
            cai_rules.append(rule)
        else:
            if rule.is_blocking:
                blocking_rules.append(rule)
            else:
                nonblocking_rules.append(rule)

    # Split up matches into respective categories
    blocking_matches_by_rule: RuleMatchMap = {}
    nonblocking_matches_by_rule: RuleMatchMap = {}
    cai_matches_by_rule: RuleMatchMap = {}
    for rule, matches in filtered_matches_by_rule.items():
        if "r2c-internal-cai" in rule.id:
            cai_matches_by_rule[rule] = [
                match for match in matches if not match.is_ignored
            ]
        else:
            if rule.is_blocking:
                blocking_matches_by_rule[rule] = [
                    match for match in matches if not match.is_ignored
                ]
            else:
                nonblocking_matches_by_rule[rule] = [
                    match for match in matches if not match.is_ignored
                ]

    sum(len(v) for v in cai_matches_by_rule.values())
    num_nonblocking_findings = sum(len(v) for v in nonblocking_matches_by_rule.values())
    num_blocking_findings = sum(len(v) for v in blocking_matches_by_rule.values())

    output_handler.output(
        blocking_matches_by_rule,
        all_targets=all_targets,
        ignore_log=ignore_log,
        profiler=profiler,
        filtered_rules=filtered_rules,
        profiling_data=profiling_data,
        severities=shown_severities,
    )
    logger.info(
        f"Ran {len(blocking_rules)} blocking rules, {len(nonblocking_rules)} audit rules, and {len(cai_rules)} internal rules used for rule recommendations."
    )
    if num_nonblocking_findings:
        logger.info(
            f"{num_nonblocking_findings} findings were from rules in audit rule board. These non-blocking findings are not displayed."
        )

    if scan_handler:
        logger.info("Reporting findings to semgrep.dev ...")
        scan_handler.report_findings(
            filtered_matches_by_rule,
            semgrep_errors,
            filtered_rules,
            all_targets,
            total_time,
            metadata.commit_datetime,
        )
        logger.info(f"Success.")

    audit_mode = metadata.event_name in audit_on
    if num_blocking_findings > 0:
        if audit_mode:
            logger.info(
                f"Audit mode is on for {metadata.event_name}, so exiting with code 0 even if matches found",
            )
            exit_code = 0
        else:
            logger.info("Has findings for blocking rules so exiting with code 1")
            exit_code = 1
    else:
        logger.info("No findings so exiting with code 0")
        exit_code = 0

    if enable_version_check:
        from semgrep.app.version import version_check

        version_check()

    sys.exit(exit_code)



#!/usr/bin/env python3
import os
import subprocess

import click

from semgrep.app import app_session
from semgrep.commands.ci import ci
from semgrep.commands.install import install_deep_semgrep
from semgrep.commands.login import login
from semgrep.commands.login import logout
from semgrep.commands.publish import publish
from semgrep.commands.scan import scan
from semgrep.default_group import DefaultGroup
from semgrep.git import GIT_SH_TIMEOUT
from semgrep.verbose_logging import getLogger

logger = getLogger(__name__)


@click.group(cls=DefaultGroup, default_command="scan", name="semgrep")
@click.help_option("--help", "-h")
@click.pass_context
def cli(ctx: click.Context) -> None:
    """
    To get started quickly, run `semgrep scan --config auto`
    """
    app_session.authenticate()
    app_session.user_agent.tags.add(
        f"command/{ctx.invoked_subcommand}"
        if ctx.invoked_subcommand in ctx.command.commands  # type: ignore
        else "command/unset"
    )

    workspace = os.getenv("GITHUB_WORKSPACE")
    if workspace:
        try:
            output = subprocess.run(
                ["git", "config", "--global", "--add", "safe.directory", workspace],
                check=True,
                encoding="utf-8",
                capture_output=True,
                timeout=GIT_SH_TIMEOUT,
            )
        except Exception as e:
            logger.info(
                f"Git safe directory workaround failed. Git commands might fail: {e}"
            )


cli.add_command(ci)
cli.add_command(login)
cli.add_command(logout)
cli.add_command(publish)
cli.add_command(scan)
cli.add_command(install_deep_semgrep)
